"""Generate a query with shredder mitigation."""

import os
import re
import sys
from datetime import date
from datetime import datetime as dt
from datetime import time, timedelta
from enum import Enum
from pathlib import Path
from types import NoneType
from typing import Any, Optional, Tuple

import attrs
import click
from dateutil import parser
from gcloud.exceptions import NotFound  # type: ignore
from google.cloud import bigquery
from jinja2 import Environment, FileSystemLoader

from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.metadata.parse_metadata import METADATA_FILE, Metadata
from bigquery_etl.metadata.validate_metadata import SHREDDER_MITIGATION_LABEL
from bigquery_etl.schema import Schema
from bigquery_etl.util.common import extract_last_group_by_from_query, write_sql

PREVIOUS_DATE = (dt.now() - timedelta(days=2)).date()
TEMP_DATASET = "tmp"
THIS_PATH = Path(os.path.dirname(__file__))
DEFAULT_PROJECT_ID = "moz-fx-data-shared-prod"
SHREDDER_MITIGATION_QUERY_NAME = "shredder_mitigation_query"
SHREDDER_MITIGATION_CHECKS_NAME = "shredder_mitigation_checks"
DEFAULT_FOR_NULLS = "??"
WILDCARD_STRING = "???????"
WILDCARD_NUMBER = -9999999
QUERY_FILE_RE = re.compile(
    r"^.*/([a-zA-Z0-9-]+)/([a-zA-Z0-9_]+)/([a-zA-Z0-9_]+(_v[0-9]+)?)/"
    r"(?:query\.sql|shredder_mitigation_query\.sql|part1\.sql|script\.sql|"
    r"query\.py|view\.sql|metadata\.yaml|backfill\.yaml)$"
)


class ColumnType(Enum):
    """Differentiate metric and dimensions."""

    METRIC = "METRIC"
    DIMENSION = "DIMENSION"
    UNDETERMINED = "None"


class ColumnStatus(Enum):
    """Different status of a column during shredder mitigation."""

    COMMON = "COMMON"
    ADDED = "ADDED"
    REMOVED = "REMOVED"
    UNDETERMINED = "None"


class DataTypeGroup(Enum):
    """Data types in BigQuery. Not supported/expected in aggregates: TIMESTAMP, ARRAY, STRUCT."""

    STRING = ("STRING", "BYTES")
    BOOLEAN = "BOOLEAN"
    INTEGER = ("INTEGER", "INT64", "INT", "SMALLINT", "TINYINT", "BYTEINT")
    NUMERIC = (
        "NUMERIC",
        "BIGNUMERIC",
    )
    FLOAT = "FLOAT"
    DATE = "DATE"
    DATETIME = "DATETIME"
    TIME = "TIME"
    TIMESTAMP = "TIMESTAMP"
    UNDETERMINED = "None"

    @staticmethod
    def from_schema_type(schema_type):
        """Convert schema type (string) to the corresponding Type enum."""
        for _type in DataTypeGroup:
            if schema_type in _type.value:
                return _type
        return None


@attrs.define(eq=True)
class Column:
    """Representation of a column in a query, with relevant details for shredder mitigation."""

    name: str
    data_type: DataTypeGroup = attrs.field(
        default=DataTypeGroup.UNDETERMINED,
        validator=attrs.validators.instance_of(DataTypeGroup),
    )
    column_type: ColumnType = attrs.field(
        default=ColumnType.UNDETERMINED,
        validator=attrs.validators.instance_of(ColumnType),
    )
    status: ColumnStatus = attrs.field(
        default=ColumnStatus.UNDETERMINED,
        validator=attrs.validators.instance_of(ColumnStatus),
    )


@attrs.define(eq=True)
class Subset:
    """Representation of a subset/CTEs in the query and the actions related to this subset."""

    @staticmethod
    def attr_not_null(instance, attribute, value):
        """Raise an exception if the value is None or empty."""
        if value is None or value == "":
            raise click.ClickException(
                f"{attribute.name} not given and it's required to continue."
            )

    client: bigquery.Client
    destination_table: str = attrs.field(
        default="",
        validator=attr_not_null,
    )
    query_cte: str = attrs.field(default="")
    dataset: str = attrs.field(default=TEMP_DATASET)
    project_id: str = attrs.field(default=DEFAULT_PROJECT_ID)
    expiration_days: Optional[float] = attrs.field(default=None)

    @property
    def version(self):
        """Return the version of the destination table."""
        match = re.search(r"v(\d+)$", self.destination_table)
        try:
            version = int(match.group(1))
            if not isinstance(version, int):
                raise click.ClickException(
                    f"{self.destination_table} must end with a positive integer."
                )
            return version
        except (AttributeError, TypeError) as e:
            raise click.ClickException(
                f"Invalid or missing table version in {self.destination_table}."
            ) from e

    @property
    def full_table_id(self):
        """Return the full id of the destination table."""
        return f"{self.project_id}.{self.dataset}.{self.destination_table}"

    @property
    def query_path(self):
        """Return the full path of the query.sql file associated with the subset."""
        sql_path = (
            Path("sql")
            / self.project_id
            / self.dataset
            / self.destination_table
            / "query.sql"
        )
        match = QUERY_FILE_RE.match(str(sql_path))
        if match and os.path.isfile(sql_path):
            return sql_path
        click.echo(click.style(f"File not found: {sql_path}.", fg="yellow"))
        return None

    @property
    def partitioning(self):
        """Return the partition details of the destination table."""
        metadata = Metadata.from_file(
            Path("sql")
            / self.project_id
            / self.dataset
            / self.destination_table
            / METADATA_FILE
        )
        if metadata.bigquery and metadata.bigquery.time_partitioning:
            partitioning = {
                "type": metadata.bigquery.time_partitioning.type.name,
                "field": metadata.bigquery.time_partitioning.field,
            }
        else:
            partitioning = {"type": None, "field": None}
        return partitioning

    @property
    def labels(self):
        """Return the labels in the metadata of the destination table."""
        metadata = Metadata.from_file(
            Path("sql")
            / self.project_id
            / self.dataset
            / self.destination_table
            / METADATA_FILE
        )
        return metadata.labels

    def generate_query(
        self,
        select_list,
        from_clause,
        where_clause=None,
        group_by_clause=None,
        order_by_clause=None,
        having_clause=None,
    ):
        """Build query to populate the table."""
        if not select_list or not from_clause:
            raise click.ClickException(
                f"Missing required clause to generate query.\n"
                f"Actual: SELECT: {select_list}, FROM: {self.full_table_id}"
            )
        query = f"SELECT {', '.join(map(str, select_list))}"
        query += f" FROM {from_clause}" if from_clause is not None else ""
        query += f" WHERE {where_clause}" if where_clause is not None else ""
        query += f" GROUP BY {group_by_clause}" if group_by_clause is not None else ""
        query += (
            f" HAVING {having_clause}"
            if having_clause is not None and group_by_clause is not None
            else ""
        )
        query += f" ORDER BY {order_by_clause}" if order_by_clause is not None else ""
        return query

    def get_query_path_results(
        self,
        backfill_date: date = PREVIOUS_DATE,
        row_limit: Optional[int] = None,
        **kwargs,
    ) -> list[dict[str, Any]]:
        """Run the query in sql_path & return result or number of rows requested."""
        having_clause = None
        for key, value in kwargs.items():
            if key.lower() == "having_clause":
                having_clause = f"{value}"

        with open(self.query_path, "r") as file:
            sql_text = file.read().strip()

        if sql_text.endswith(";"):
            sql_text = sql_text[:-1]
        if having_clause:
            sql_text = f"{sql_text} {having_clause}"
        if row_limit:
            sql_text = f"{sql_text} LIMIT {row_limit}"

        partition_field = self.partitioning["field"]
        partition_type = (
            "DATE" if self.partitioning["type"] == "DAY" else self.partitioning["type"]
        )
        parameters = []
        if partition_field is not None:
            parameters = [
                bigquery.ScalarQueryParameter(
                    partition_field, partition_type, backfill_date
                ),
            ]

        try:
            query_job = self.client.query(
                query=sql_text,
                job_config=bigquery.QueryJobConfig(
                    query_parameters=parameters,
                    use_legacy_sql=False,
                    dry_run=False,
                    use_query_cache=False,
                ),
            )
            query_results = query_job.result()
        except NotFound as e:
            raise click.ClickException(
                f"Unable to query data for {backfill_date}. Table {self.full_table_id} not found"
                f" or partitioning in metadata missing."
            ) from e
        rows = [dict(row) for row in query_results]
        return rows


def get_bigquery_type(value) -> DataTypeGroup:
    """Find the datatype of a value, grouping similar types."""
    if isinstance(value, dt):
        return DataTypeGroup.DATETIME
    try:
        if isinstance(dt.strptime(value, "%H:%M:%S").time(), time):
            return DataTypeGroup.TIME
    except (ValueError, TypeError, AttributeError):
        pass
    try:
        value_parsed = parser.isoparse(value.replace(" UTC", "Z").replace(" ", "T"))
        if (
            isinstance(value_parsed, dt)
            and value_parsed.time() == time(0, 0)
            and isinstance(dt.strptime(value, "%Y-%m-%d"), date)
        ):
            return DataTypeGroup.DATE
        if isinstance(value_parsed, dt) and value_parsed.tzinfo is None:
            return DataTypeGroup.DATETIME
        if isinstance(value_parsed, dt) and value_parsed.tzinfo is not None:
            return DataTypeGroup.TIMESTAMP
    except (ValueError, TypeError, AttributeError):
        pass
    if isinstance(value, time):
        return DataTypeGroup.TIME
    if isinstance(value, date):
        return DataTypeGroup.DATE
    if isinstance(value, bool):
        return DataTypeGroup.BOOLEAN
    if isinstance(value, int):
        return DataTypeGroup.INTEGER
    if isinstance(value, float):
        return DataTypeGroup.FLOAT
    if isinstance(value, (str, bytes)):
        return DataTypeGroup.STRING
    if isinstance(value, NoneType):
        return DataTypeGroup.UNDETERMINED
    raise ValueError(f"Unsupported data type: {type(value)}")


def validate_types(columns, schema, sample_data):
    """Return dtype from schema file or if not available, from bigquery."""
    results = {}

    for dim in columns:
        schema_field = next((field for field in schema if field.name == dim), None)

        bigquery_type = get_bigquery_type(sample_data.get(dim))
        # If the column is in the schema
        if schema_field:
            schema_type = DataTypeGroup.from_schema_type(schema_field.field_type)
            results[dim] = schema_type or bigquery_type
        # If the column is _not_ in the schema
        else:
            if bigquery_type:
                results[dim] = bigquery_type
                click.echo(
                    f"Column {dim} not in schema!! Using type {bigquery_type} from sample data."
                )
            else:
                results[dim] = DataTypeGroup.UNDETERMINED
                click.echo(
                    f"Data type for {dim} could not be retrieved from schema or BigQuery."
                    f" Set to UNDETERMINED."
                )
                continue
    return results


def classify_columns(
    new_row: dict,
    existing_dimension_columns: list,
    new_dimension_columns: list,
    existing_schema: bigquery.schema.SchemaField,
    new_schema: bigquery.schema.SchemaField,
) -> tuple[list[Column], list[Column], list[Column], list[Column], list[Column]]:
    """Compare new row with existing columns & return common, added & removed columns."""
    common_dimensions = []
    added_dimensions = []
    removed_dimensions = []
    metrics = []
    undefined = []

    if not new_row or not existing_dimension_columns or not new_dimension_columns:
        raise click.ClickException(
            f"\n\nMissing one or more required parameters. Received:"
            f"\nnew_row= {new_row}"
            f"\nexisting_dimension_columns= {existing_dimension_columns},"
            f"\nnew_dimension_columns= {new_dimension_columns}."
        )

    missing_dimensions = [
        dimension for dimension in new_dimension_columns if dimension not in new_row
    ]
    if not len(missing_dimensions) == 0:
        raise click.ClickException(
            f"Existing dimensions don't match columns retrieved by query."
            f" Missing {missing_dimensions}."
        )
    existing_dimension_columns_types = validate_types(
        existing_dimension_columns, existing_schema, new_row
    )
    new_dimension_columns_types = validate_types(
        new_dimension_columns, new_schema, new_row
    )

    for key in existing_dimension_columns:
        if key not in new_dimension_columns:
            removed_dimensions.append(
                Column(
                    key,
                    existing_dimension_columns_types[key],
                    ColumnType.UNDETERMINED,
                    ColumnStatus.REMOVED,
                )
            )

    for key, _ in new_row.items():
        if key in existing_dimension_columns:
            common_dimensions.append(
                Column(
                    key,
                    new_dimension_columns_types[key],
                    ColumnType.DIMENSION,
                    ColumnStatus.COMMON,
                )
            )
        elif key not in existing_dimension_columns and key in new_dimension_columns:
            added_dimensions.append(
                Column(
                    key,
                    new_dimension_columns_types[key],
                    ColumnType.DIMENSION,
                    ColumnStatus.ADDED,
                )
            )
        elif key not in existing_dimension_columns and key not in new_dimension_columns:
            value_type = validate_types([key], new_schema, new_row)
            if value_type[key] in (
                DataTypeGroup.INTEGER,
                DataTypeGroup.FLOAT,
                DataTypeGroup.NUMERIC,
            ):
                # A column that's not a dimensions nor a grouping columns is classified as metric.
                metrics.append(
                    Column(
                        key,
                        value_type[key],
                        ColumnType.METRIC,
                        ColumnStatus.COMMON,
                    )
                )
            else:
                undefined.append(
                    Column(
                        key,
                        value_type[key],
                        ColumnType.UNDETERMINED,
                        ColumnStatus.UNDETERMINED,
                    )
                )

    common_dimensions_sorted = sorted(common_dimensions, key=lambda column: column.name)
    added_dimensions_sorted = sorted(added_dimensions, key=lambda column: column.name)
    removed_dimensions_sorted = sorted(
        removed_dimensions, key=lambda column: column.name
    )
    metrics_sorted = sorted(metrics, key=lambda column: column.name)
    undefined_sorted = sorted(undefined, key=lambda column: column.name)

    return (
        common_dimensions_sorted,
        added_dimensions_sorted,
        removed_dimensions_sorted,
        metrics_sorted,
        undefined_sorted,
    )


def generate_query_with_shredder_mitigation(
    client,
    project_id,
    dataset,
    destination_table,
    staging_table_name,
    backfill_date=PREVIOUS_DATE,
) -> Tuple[Path, str]:
    """Generate a query to backfill with shredder mitigation."""
    query_with_mitigation_path = Path("sql") / project_id

    # Find query files and grouping of previous and new queries.
    new = Subset(client, destination_table, "new_version", dataset, project_id, None)

    if SHREDDER_MITIGATION_LABEL not in new.labels:
        click.echo(
            click.style(
                "The required label `shredder_mitigation` is missing in the metadata of the "
                "table. The process will now terminate.",
                fg="yellow",
            )
        )
        sys.exit(1)

    if new.version < 2:
        raise click.ClickException(
            f"The new version of the table is expected >= 2. Actual is {new.version}."
        )

    destination_table_previous_version = (
        f"{destination_table[:-len(str(new.version))]}{new.version-1}"
    )
    previous = Subset(
        client,
        destination_table_previous_version,
        "previous",
        dataset,
        project_id,
        None,
    )
    new_group_by = extract_last_group_by_from_query(sql_path=new.query_path)
    previous_group_by = extract_last_group_by_from_query(sql_path=previous.query_path)

    # Check that previous query exists and GROUP BYs are valid in both queries.
    integers_in_group_by = False
    for e in previous_group_by + new_group_by:
        try:
            int(e)
            integers_in_group_by = True
        except ValueError:
            continue
    if (
        "ALL" in previous_group_by
        or "ALL" in new_group_by
        or not all(isinstance(e, str) for e in previous_group_by)
        or not all(isinstance(e, str) for e in new_group_by)
        or not previous_group_by
        or not new_group_by
        or integers_in_group_by
    ):
        raise click.ClickException(
            "GROUP BY must use an explicit list of columns. "
            "Avoid expressions like `GROUP BY ALL` or `GROUP BY 1, 2, 3`."
        )

    # Identify columns common to both queries and columns new. This excludes removed columns.
    existing_schema = Schema.from_schema_file(
        query_with_mitigation_path
        / dataset
        / destination_table_previous_version
        / "schema.yaml"
    ).to_bigquery_schema()
    new_schema = Schema.from_schema_file(
        query_with_mitigation_path / dataset / destination_table / "schema.yaml"
    ).to_bigquery_schema()

    sample_rows = new.get_query_path_results(
        backfill_date=backfill_date,
        row_limit=1,
        having_clause=f"HAVING {' IS NOT NULL AND '.join(new_group_by)} IS NOT NULL",
    )
    if not sample_rows:
        sample_rows = new.get_query_path_results(
            backfill_date=backfill_date, row_limit=1
        )
    try:
        new_table_row = sample_rows[0]
        (
            common_dimensions,
            added_dimensions,
            removed_dimensions,
            metrics,
            undetermined_columns,
        ) = classify_columns(
            new_table_row, previous_group_by, new_group_by, existing_schema, new_schema
        )
    except TypeError:
        raise click.ClickException(
            f"Table {destination_table} did not return any rows for {backfill_date}."
        )

    if not common_dimensions or not added_dimensions or not metrics:
        raise click.ClickException(
            "The process requires that previous & query have at least one dimension in common,"
            " one dimension added and one metric."
        )

    # Find the new version of the query.
    with open(new.query_path, "r") as file:
        new_query = file.read().strip()

    # Aggregate previous data and new query results using common dimensions.
    new_agg = Subset(
        client, destination_table, "new_agg", TEMP_DATASET, project_id, None
    )
    previous_agg = Subset(
        client,
        destination_table_previous_version,
        "previous_agg",
        TEMP_DATASET,
        project_id,
        None,
    )
    common_select = (
        [previous.partitioning["field"]]
        + [
            f"IF({dim.name} IS NULL OR {dim.name} = '{DEFAULT_FOR_NULLS}', '{WILDCARD_STRING}',"
            f" {dim.name}) AS {dim.name}"
            for dim in common_dimensions
            if (
                dim.name != previous.partitioning["field"]
                and dim.data_type == DataTypeGroup.STRING
            )
        ]
        + [
            f"COALESCE({dim.name}, {WILDCARD_NUMBER}) AS {dim.name}"
            for dim in common_dimensions
            if (
                dim.name != new.partitioning["field"]
                and dim.data_type in (DataTypeGroup.INTEGER, DataTypeGroup.FLOAT)
            )
        ]
        + [
            dim.name
            for dim in common_dimensions
            if (
                dim.name != new.partitioning["field"]
                and dim.data_type in (DataTypeGroup.BOOLEAN, DataTypeGroup.DATE)
            )
        ]
        + [f"SUM({metric.name}) AS {metric.name}" for metric in metrics]
    )
    new_agg_query = new_agg.generate_query(
        select_list=common_select,
        from_clause=f"{new.query_cte}",
        group_by_clause="ALL",
    )
    previous_agg_query = previous_agg.generate_query(
        select_list=common_select,
        from_clause=f"`{previous.full_table_id}`",
        where_clause=f"{previous.partitioning['field']} = @{previous.partitioning['field']}",
        group_by_clause="ALL",
    )

    # Generate the subset that will calculate shredder impact.
    shredded = Subset(
        client, destination_table, "shredded", TEMP_DATASET, project_id, None
    )

    # Cast NULL values to the corresponding type.
    # This doesn't convert data or dtypes, it's only used to cast NULLs for UNION queries.
    shredded_select = (
        [f"{previous_agg.query_cte}.{new.partitioning['field']}"]
        + [
            f"{previous_agg.query_cte}.{dim.name}"
            for dim in common_dimensions
            if (dim.name != new.partitioning["field"])
        ]
        + [
            f"CAST(NULL AS {DataTypeGroup.STRING.name}) AS {dim.name}"
            for dim in added_dimensions
            if (
                dim.name != new.partitioning["field"]
                and dim.data_type == DataTypeGroup.STRING
            )
        ]
        + [
            f"CAST(NULL AS {DataTypeGroup.BOOLEAN.name}) AS {dim.name}"
            for dim in added_dimensions
            if (
                dim.name != new.partitioning["field"]
                and dim.data_type == DataTypeGroup.BOOLEAN
            )
        ]
        + [
            f"CAST(NULL AS {DataTypeGroup.DATE.name}) AS {dim.name}"
            for dim in added_dimensions
            if (
                dim.name != new.partitioning["field"]
                and dim.data_type == DataTypeGroup.DATE
            )
        ]
        + [
            f"CAST(NULL AS {DataTypeGroup.INTEGER.name}) AS {dim.name}"
            for dim in added_dimensions
            if (
                dim.name != new.partitioning["field"]
                and dim.data_type == DataTypeGroup.INTEGER
            )
        ]
        + [
            f"CAST(NULL AS {DataTypeGroup.FLOAT.name}) AS {dim.name}"
            for dim in added_dimensions
            if (
                dim.name != new.partitioning["field"]
                and dim.data_type == DataTypeGroup.FLOAT
            )
        ]
        + [
            f"NULL AS {dim.name}"
            for dim in added_dimensions
            if (
                dim.name != new.partitioning["field"]
                and dim.data_type
                not in (
                    DataTypeGroup.STRING,
                    DataTypeGroup.BOOLEAN,
                    DataTypeGroup.INTEGER,
                    DataTypeGroup.FLOAT,
                )
            )
        ]
        + [
            f"COALESCE({previous_agg.query_cte}.{metric.name}, 0) - "
            f"COALESCE({new_agg.query_cte}.{metric.name}, 0)"
            f" AS {metric.name}"
            for metric in metrics
            if metric.data_type != DataTypeGroup.FLOAT
        ]
        + [
            # Round FLOAT to avoid exponential numbers.
            f"ROUND({previous_agg.query_cte}.{metric.name}, 10) - "
            f"ROUND(COALESCE({new_agg.query_cte}.{metric.name}, 0), 10) AS {metric.name}"
            for metric in metrics
            if metric.data_type == DataTypeGroup.FLOAT
        ]
    )

    shredded_join = " AND ".join(
        [
            f"{previous_agg.query_cte}.{previous.partitioning['field']} ="
            f" {new_agg.query_cte}.{new.partitioning['field']}"
        ]
        + [
            f"{previous_agg.query_cte}.{dim.name} = {new_agg.query_cte}.{dim.name}"
            for dim in common_dimensions
            if dim.name != previous.partitioning["field"]
            and dim.data_type not in (DataTypeGroup.BOOLEAN, DataTypeGroup.DATE)
        ]
        + [
            f"({previous_agg.query_cte}.{dim.name} = {new_agg.query_cte}.{dim.name} OR"
            f" ({previous_agg.query_cte}.{dim.name} IS NULL"
            f" AND {new_agg.query_cte}.{dim.name} IS NULL))"  # Compare null values.
            for dim in common_dimensions
            if dim.name != previous.partitioning["field"]
            and dim.data_type in (DataTypeGroup.BOOLEAN, DataTypeGroup.DATE)
        ]
    )
    shredded_query = shredded.generate_query(
        select_list=shredded_select,
        from_clause=f"{previous_agg.query_cte} LEFT JOIN {new_agg.query_cte} ON {shredded_join} ",
        where_clause=" OR ".join(
            [
                f"COALESCE({previous_agg.query_cte}.{metric.name}, 0) >"
                f" COALESCE({new_agg.query_cte}.{metric.name}, 0)"
                for metric in metrics
            ]
        ),
    )

    combined_list = []

    # Set back to NULL, values that were temporarily set to a wildcard.
    for dim in common_dimensions + added_dimensions:
        if dim.data_type.name == DataTypeGroup.STRING.name:
            combined_list.append(
                f"IF({dim.name} = '{WILDCARD_STRING}', "
                f"CAST(NULL AS {DataTypeGroup.STRING.name}), {dim.name}) AS {dim.name}"
            )
        elif dim.data_type.name == DataTypeGroup.INTEGER.name:
            combined_list.append(
                f"IF({dim.name} = {WILDCARD_NUMBER}, "
                f"CAST(NULL AS {DataTypeGroup.INTEGER.name}), {dim.name}) AS {dim.name}"
            )
        elif dim.data_type.name == DataTypeGroup.NUMERIC.name:
            combined_list.append(
                f"IF({dim.name} = '{WILDCARD_STRING}', "
                f"CAST(NULL AS {DataTypeGroup.NUMERIC.name}), {dim.name}) AS {dim.name}"
            )
        elif dim.data_type.name == DataTypeGroup.FLOAT.name:
            combined_list.append(
                f"IF({dim.name} = '{WILDCARD_STRING}', "
                f"CAST(NULL AS {DataTypeGroup.FLOAT.name}), {dim.name}) AS {dim.name}"
            )
        else:
            combined_list.append(f"{dim.name} AS {dim.name}")

    combined_list = combined_list + [metric.name for metric in metrics]
    final_select = f"{', '.join(combined_list)}"

    # Generate formatted output strings to display generated-query information in console.
    common_output = "".join(
        [
            f"{dim.column_type.name} > {dim.name}:{dim.data_type.name}\n"
            for dim in common_dimensions
        ]
    )
    metrics_output = "".join(
        [
            f"{dim.column_type.name} > {dim.name}:{dim.data_type.name}\n"
            for dim in metrics
        ]
    )
    changed_output = "".join(
        [
            f"{dim.status.name} > {dim.name}:{dim.data_type.name}\n"
            for dim in added_dimensions + removed_dimensions + undetermined_columns
        ]
    )
    click.echo(
        click.style(
            f"Query columns:\n" f"{common_output + metrics_output + changed_output}",
            fg="yellow",
        )
    )

    # Generate query using the template.
    env = Environment(loader=FileSystemLoader(str(THIS_PATH)))
    query_with_mitigation_template = env.get_template(
        f"{SHREDDER_MITIGATION_QUERY_NAME}_template.sql"
    )
    checks_for_mitigation_template = env.get_template(
        f"{SHREDDER_MITIGATION_CHECKS_NAME}_template.sql"
    )

    query_with_mitigation_sql = reformat(
        query_with_mitigation_template.render(
            new_version_cte=new.query_cte,
            new_version=new_query,
            new_agg_cte=new_agg.query_cte,
            new_agg=new_agg_query,
            previous_agg_cte=previous_agg.query_cte,
            previous_agg=previous_agg_query,
            shredded_cte=shredded.query_cte,
            shredded=shredded_query,
            final_select=final_select,
        )
    )
    write_sql(
        output_dir=query_with_mitigation_path,
        full_table_id=new.full_table_id,
        basename=f"{SHREDDER_MITIGATION_QUERY_NAME}.sql",
        sql=query_with_mitigation_sql,
        skip_existing=False,
    )

    # Generate checks to compare versions after each partition backfill.
    checks_select = (
        [new.partitioning["field"]]
        + [
            f"IF({dim.name} IS NULL OR {dim.name} = '{DEFAULT_FOR_NULLS}', '{WILDCARD_STRING}',"
            f" {dim.name}) AS {dim.name}"
            for dim in common_dimensions
            if (
                dim.name != previous.partitioning["field"]
                and dim.data_type == DataTypeGroup.STRING
            )
        ]
        + [
            dim.name
            for dim in common_dimensions
            if (
                dim.name != new.partitioning["field"]
                and dim.data_type != DataTypeGroup.STRING
            )
        ]
        + [f"SUM({metric.name})" f" AS {metric.name}" for metric in metrics]
    )
    previous_checks_query = previous.generate_query(
        select_list=checks_select,
        from_clause=f"`{previous.full_table_id}`",
        where_clause=f"{previous.partitioning['field']} = @{previous.partitioning['field']}",
        group_by_clause="ALL",
    )
    new_checks_query = new.generate_query(
        select_list=checks_select,
        from_clause=f"`{staging_table_name}`",
        where_clause=f"{previous.partitioning['field']} = @{previous.partitioning['field']}",
        group_by_clause="ALL",
    )

    checks_for_mitigation_sql = reformat(
        checks_for_mitigation_template.render(
            previous_version_cte=previous.query_cte,
            previous_version=previous_checks_query,
            new_version_cte=new.query_cte,
            new_version=new_checks_query,
        )
    )
    write_sql(
        output_dir=query_with_mitigation_path,
        full_table_id=new.full_table_id,
        basename=f"{SHREDDER_MITIGATION_CHECKS_NAME}.sql",
        sql=checks_for_mitigation_sql,
        skip_existing=False,
    )

    query_path = Path("sql") / new.project_id / new.dataset / new.destination_table

    click.echo(
        click.style(
            f"Files for shredder mitigation written to path `{query_path}`.\n"
            f"Query: `{SHREDDER_MITIGATION_QUERY_NAME}.sql`\n"
            f"Data checks: `{SHREDDER_MITIGATION_CHECKS_NAME}.sql`\n",
            fg="blue",
        )
    )

    return (
        query_path,
        query_with_mitigation_sql,
    )



import os
from pathlib import Path

import pytest
from generate_queries import Template, TemplatedDir

BASE_DIR = Path(os.path.dirname(__file__)).parent


class TestTemplatedDir:
    @pytest.fixture
    def query_dir(self):
        return TemplatedDir("event_types", Path(BASE_DIR / "templates" / "event_types"))

    def test_get_datasets(self, query_dir):
        assert query_dir.get_datasets(query_dir.get_args()) == [
            "dataset-1",
            "dataset-2",
        ]

    def test_get_datasets_with_arg(self, query_dir):
        assert query_dir.get_datasets(query_dir.get_args(), "dataset-2") == [
            "dataset-2"
        ]

    def test_get_templates(self, query_dir):
        assert query_dir.get_templates() == [
            Template("query.sql", query_dir.get_environment())
        ]

    def test_get_args(self, query_dir):
        assert query_dir.get_args() == {
            "dataset-1": {"key": "val1"},
            "dataset-2": {"key": "val2"},
        }



"""bqetl_project.yaml config loader."""

from pathlib import Path

import yaml

ROOT = Path(__file__).parent.parent
BQETL_PROJECT_CONFIG = "bqetl_project.yaml"


class _ConfigLoader:
    project_dir: Path = ROOT
    config_file: str = BQETL_PROJECT_CONFIG

    @property
    def config(self):
        config = getattr(self, "_config", None)
        if config:
            return config

        self._config = yaml.safe_load((self.project_dir / self.config_file).read_text())
        return self._config

    def set_project_dir(self, project_dir: Path):
        """Update the project root directory."""
        self.project_dir = project_dir

    def get(self, *args, fallback=None):
        """Get the config option specified by args."""
        conf = self.config

        for arg in args:
            if arg in conf:
                conf = conf[arg]
            else:
                return fallback

        return conf


ConfigLoader = _ConfigLoader()



"""Represents a SQL view."""

import glob
import re
import string
import sys
import time
from functools import cached_property
from pathlib import Path
from textwrap import dedent
from typing import Any, Optional

import attr
import sqlparse
from google.api_core.exceptions import BadRequest, NotFound
from google.cloud import bigquery

from bigquery_etl.config import ConfigLoader
from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.metadata.parse_metadata import (
    DATASET_METADATA_FILE,
    METADATA_FILE,
    DatasetMetadata,
    Metadata,
)
from bigquery_etl.schema import SCHEMA_FILE, Schema
from bigquery_etl.util import extract_from_query_path
from bigquery_etl.util.common import render

# Regex matching CREATE VIEW statement so it can be removed to get the view query
CREATE_VIEW_PATTERN = re.compile(
    r"CREATE\s+OR\s+REPLACE\s+VIEW\s+[^\s]+\s+AS", re.IGNORECASE
)


@attr.s(auto_attribs=True)
class View:
    """Representation of a SQL view stored in a view.sql file."""

    path: str = attr.ib()
    name: str = attr.ib()
    dataset: str = attr.ib()
    project: str = attr.ib()
    partition_column: Optional[str] = attr.ib(None)
    id_token: Optional[Any] = attr.ib(None)

    @path.validator
    def validate_path(self, attribute, value):
        """Check that the view path is valid."""
        if not Path(self.path).exists():
            raise ValueError(f"View file does not exist: {self.path}")

    @property
    def content(self):
        """Return the view SQL."""
        path = Path(self.path)
        return render(path.name, template_folder=path.parent)

    @classmethod
    def from_file(cls, path, **kwargs):
        """View from SQL file."""
        project, dataset, name = extract_from_query_path(path)
        return cls(
            path=str(path), name=name, dataset=dataset, project=project, **kwargs
        )

    @property
    def view_identifier(self):
        """Return full view identifier: `<project>.<dataset>.<name>`."""
        return f"{self.project}.{self.dataset}.{self.name}"

    @property
    def is_user_facing(self):
        """Return whether the view is user-facing."""
        return not self.dataset.endswith(
            tuple(
                ConfigLoader.get(
                    "default", "non_user_facing_dataset_suffixes", fallback=[]
                )
            )
        )

    @cached_property
    def metadata(self):
        """Return the view metadata."""
        path = Path(self.path).parent / METADATA_FILE
        if not path.exists():
            return None
        return Metadata.from_file(path)

    @property
    def labels(self):
        """Return the view labels."""
        if not hasattr(self, "_labels"):
            if self.metadata:
                self._labels = self.metadata.labels.copy()
            else:
                self._labels = {}
        return self._labels

    @classmethod
    def create(cls, project, dataset, name, sql_dir, base_table=None):
        """
        Create a new empty view from a template.

        Use `base_table` in view definition, if provided.
        """
        path = Path(sql_dir) / project / dataset / name / "view.sql"
        dataset_path = path.parent.parent

        if not dataset_path.exists():
            # create new dataset with dataset metadata
            path.parent.mkdir(parents=True)
            dataset_metadata = DatasetMetadata(
                friendly_name=string.capwords(dataset),
                description="Please provide a dataset description.",
                dataset_base_acl="view",
                user_facing=True,
            )
            dataset_metadata.write(dataset_path / DATASET_METADATA_FILE)
        else:
            path.parent.mkdir(parents=True, exist_ok=True)

        if not base_table:
            base_table = f"{project}.{dataset}_derived.{name}_v1"

        path.write_text(
            reformat(
                f"""
                CREATE OR REPLACE VIEW `{project}.{dataset}.{name}` AS
                SELECT * FROM `{base_table}`
                """
            )
            + "\n"
        )
        return cls(path, name, dataset, project)

    def skip_validation(self):
        """Get views that should be skipped during validation."""
        return {
            file
            for skip in ConfigLoader.get("view", "validation", "skip", fallback=[])
            for file in glob.glob(
                skip,
                recursive=True,
            )
        }

    def skip_publish(self):
        """Get views that should be skipped during publishing."""
        return {
            file
            for skip in ConfigLoader.get("view", "publish", "skip", fallback=[])
            for file in glob.glob(
                skip,
                recursive=True,
            )
        }

    def is_valid(self) -> bool:
        """Validate the SQL view definition."""
        if any(str(self.path).endswith(p) for p in self.skip_validation()):
            print(f"Skipped validation for {self.path}")
            return True
        return self._valid_fully_qualified_references() and self._valid_view_naming()

    @cached_property
    def table_references(self):
        """List of table references in this view."""
        from bigquery_etl.dependency import extract_table_references

        return extract_table_references(self.content)

    @cached_property
    def udf_references(self):
        """List of UDF references in this view."""
        from bigquery_etl.routine.parse_routine import routine_usages_in_text

        # routine_usages_in_text automatically includes mozfun UDFs
        return routine_usages_in_text(
            self.content, Path(self.path).parent.parent.parent
        )

    @cached_property
    def schema(self):
        """Derive view schema from a schema file or a dry run result."""
        return self.configured_schema or self.dryrun_schema

    @cached_property
    def schema_path(self):
        """Return the schema file path."""
        return Path(self.path).parent / SCHEMA_FILE

    @cached_property
    def configured_schema(self):
        """Derive view schema from a schema file."""
        if self.schema_path.is_file():
            return Schema.from_schema_file(self.schema_path)
        return None

    @cached_property
    def dryrun_schema(self):
        """Derive view schema from a dry run result."""
        try:
            # We have to remove `CREATE OR REPLACE VIEW ... AS` from the query to avoid
            # view-creation-permission-denied errors, and we have to apply a `WHERE`
            # filter to avoid partition-column-filter-missing errors.
            schema_query_filter = (
                f"DATE(`{self.partition_column}`) = DATE('2020-01-01')"
                if self.partition_column
                else "FALSE"
            )
            schema_query = dedent(
                f"""
                WITH view_query AS (
                    {CREATE_VIEW_PATTERN.sub("", self.content)}
                )
                SELECT *
                FROM view_query
                WHERE {schema_query_filter}
                """
            )
            return Schema.from_query_file(
                Path(self.path), content=schema_query, id_token=self.id_token
            )
        except Exception as e:
            print(f"Error dry-running view {self.view_identifier} to get schema: {e}")
            return None

    def _valid_fully_qualified_references(self):
        """Check that referenced tables and views are fully qualified."""
        for table in self.table_references:
            if len(table.split(".")) < 3:
                print(f"{self.path} ERROR\n{table} missing project_id qualifier")
                return False
        return True

    def _valid_view_naming(self):
        """Validate that the created view naming matches the directory structure."""
        if not (parsed := sqlparse.parse(self.content)):
            raise ValueError(f"Unable to parse view SQL for {self.path}")

        tokens = [
            t
            for t in parsed[0].tokens
            if not (t.is_whitespace or isinstance(t, sqlparse.sql.Comment))
        ]
        is_view_statement = (
            " ".join(tokens[0].normalized.split()) == "CREATE OR REPLACE"
            and tokens[1].normalized == "VIEW"
        )
        if is_view_statement:
            target_view = str(tokens[2]).strip().split()[0]
            try:
                [project_id, dataset_id, view_id] = target_view.replace("`", "").split(
                    "."
                )
                if not (
                    self.name == view_id
                    and self.dataset == dataset_id
                    and self.project == project_id
                ):
                    print(
                        f"{self.path} ERROR\n"
                        f"View name {target_view} not matching directory structure."
                    )
                    return False
            except Exception:
                print(f"{self.path} ERROR\n{target_view} missing project ID qualifier.")
                return False
        else:
            print(
                f"ERROR: {self.path} does not appear to be "
                "a CREATE OR REPLACE VIEW statement! Quitting..."
            )
            return False
        return True

    def target_view_identifier(self, target_project=None):
        """Return the view identifier after replacing project with target_project.

        Result must be a fully-qualified BigQuery Standard SQL table identifier, which
        is of the form f"{project_id}.{dataset_id}.{table_id}". dataset_id and table_id
        may not contain "." or "`". Each component may be a backtick (`) quoted
        identifier, or the whole thing may be a backtick quoted identifier, but not
        both. Project IDs must contain 6-63 lowercase letters, digits, or dashes. Some
        project IDs also include domain name separated by a colon. IDs must start with a
        letter and may not end with a dash. For more information see also
        https://github.com/mozilla/bigquery-etl/pull/1427#issuecomment-707376291
        """
        if target_project:
            return self.view_identifier.replace(self.project, target_project, 1)
        return self.view_identifier

    def has_changes(self, target_project=None, credentials=None):
        """Determine whether there are any changes that would be published."""
        if any(str(self.path).endswith(p) for p in self.skip_publish()):
            return False

        if target_project and self.project != ConfigLoader.get(
            "default", "project", fallback="moz-fx-data-shared-prod"
        ):
            # view would be skipped because --target-project is set
            return False

        if credentials:
            client = bigquery.Client(credentials=credentials)
        else:
            client = bigquery.Client()
        target_view_id = self.target_view_identifier(target_project)
        try:
            table = client.get_table(target_view_id)
        except NotFound:
            print(f"view {target_view_id} will change: does not exist in BigQuery")
            return True

        try:
            expected_view_query = CREATE_VIEW_PATTERN.sub(
                "", sqlparse.format(self.content, strip_comments=True), count=1
            ).strip(";" + string.whitespace)

            actual_view_query = sqlparse.format(
                table.view_query, strip_comments=True
            ).strip(";" + string.whitespace)
        except TypeError:
            print(
                f"ERROR: There has been an issue formatting: {target_view_id}",
                file=sys.stderr,
            )
            raise

        if expected_view_query != actual_view_query:
            print(f"view {target_view_id} will change: query does not match")
            return True

        # check metadata
        if self.metadata is not None:
            if self.metadata.description != table.description:
                print(f"view {target_view_id} will change: description does not match")
                return True
            if self.metadata.friendly_name != table.friendly_name:
                print(
                    f"view {target_view_id} will change: friendly_name does not match"
                )
                return True
            if self.labels != table.labels:
                print(f"view {target_view_id} will change: labels do not match")
                return True

        table_schema = Schema.from_bigquery_schema(table.schema)

        if self.schema is not None and not self.schema.equal(table_schema):
            print(f"view {target_view_id} will change: schema does not match")
            return True

        return False

    def publish(self, target_project=None, dry_run=False, client=None):
        """
        Publish this view to BigQuery.

        If `target_project` is set, it will replace the project ID in the view definition.
        """
        if any(str(self.path).endswith(p) for p in self.skip_publish()):
            print(f"Skipping {self.path}")
            return True

        # avoid checking references since Jenkins might throw an exception:
        # https://github.com/mozilla/bigquery-etl/issues/2246
        if (
            any(str(self.path).endswith(p) for p in self.skip_validation())
            or self._valid_view_naming()
        ):
            client = client or bigquery.Client()
            sql = self.content
            target_view = self.target_view_identifier(target_project)

            if target_project:
                if self.project != ConfigLoader.get(
                    "default", "project", fallback="moz-fx-data-shared-prod"
                ):
                    print(f"Skipping {self.path} because --target-project is set")
                    return True

                # We only change the first occurrence, which is in the target view name.
                sql = re.sub(
                    rf"^(?!--)(.*){self.project}",
                    rf"\1{target_project}",
                    sql,
                    count=1,
                    flags=re.MULTILINE,
                )

            job_config = bigquery.QueryJobConfig(use_legacy_sql=False, dry_run=dry_run)
            query_job = client.query(sql, job_config)

            if dry_run:
                print(f"Validated definition of {target_view} in {self.path}")
            else:
                try:
                    job_id = query_job.result().job_id
                except BadRequest as e:
                    if "Invalid snapshot time" in e.message:
                        # This occasionally happens due to dependent views being
                        # published concurrently; we wait briefly and give it one
                        # extra try in this situation.
                        time.sleep(1)
                        job_id = client.query(sql, job_config).result().job_id
                    else:
                        raise

                try:
                    table = client.get_table(target_view)
                except NotFound:
                    print(
                        f"{target_view} failed to publish to the correct location, verify job id {job_id}",
                        file=sys.stderr,
                    )
                    return False

                try:
                    if self.schema_path.is_file():
                        table = self.schema.deploy(target_view)
                except Exception as e:
                    print(f"Could not update field descriptions for {target_view}: {e}")

                if not self.metadata:
                    print(f"Missing metadata for {self.path}")

                table.description = self.metadata.description
                table.friendly_name = self.metadata.friendly_name

                if table.labels != self.labels:
                    labels = self.labels.copy()
                    for key in table.labels:
                        if key not in labels:
                            # To delete a label its value must be set to None
                            labels[key] = None
                    table.labels = {
                        key: value
                        for key, value in labels.items()
                        if isinstance(value, str)
                    }
                    client.update_table(
                        table, ["labels", "description", "friendly_name"]
                    )
                else:
                    client.update_table(table, ["description", "friendly_name"])

                print(f"Published view {target_view}")
        else:
            print(
                f"Error publishing {self.path}. Invalid view definition.",
                file=sys.stderr,
            )
            return False

        return True



#!/usr/bin/env python

"""Report estimated cost to run shredder."""

import warnings
from argparse import ArgumentParser
from datetime import datetime, timedelta
from math import ceil
from textwrap import dedent

from google.cloud import bigquery

from ..util import standard_args
from ..util.bigquery_id import sql_table_id
from .config import DELETE_TARGETS

JOBS_QUERY = """
SELECT
  SUM(total_bytes_processed) AS total_bytes_processed,
  SUM(total_slot_ms) AS slot_millis,
  destination_table.project_id AS project,
  destination_table.dataset_id,
  destination_table.table_id
FROM
  `moz-fx-data-bq-batch-prod.region-us.INFORMATION_SCHEMA.JOBS_BY_PROJECT`
WHERE
  job_id IN (
    -- most recent unqualified job id for each task
    SELECT
      SPLIT(ARRAY_AGG(job_id ORDER BY job_created DESC)[OFFSET(0)], ".")[OFFSET(1)]
    FROM
      `moz-fx-data-shredder.shredder_state.shredder_state`
    GROUP BY
      task_id
  )
  AND state = "DONE"
  AND error_result IS NULL
  AND destination_table.dataset_id = "telemetry_derived"
GROUP BY
  destination_table.project_id,
  destination_table.dataset_id,
  destination_table.table_id
"""

TABLES_QUERY = (
    "WITH tables AS ("
    + "\n  UNION ALL".join(
        f"\n  SELECT * FROM `{project}.{dataset_id}.__TABLES__`"
        for project, dataset_id in {
            (target.project, target.dataset_id) for target in DELETE_TARGETS
        }
    )
    + """
)
SELECT
  project_id AS project,
  dataset_id,
  table_id,
  size_bytes AS num_bytes,
FROM
  tables
"""
)

# calculate the minimum bytes_per_second a job needs to process to reduce cost
# by using flat rate pricing instead of on demand pricing at 100% utilization
seconds_per_day = 60 * 60 * 24
seconds_per_year = seconds_per_day * 365
flat_rate_dollars_per_year_per_slot = 12 * 8500 / 500
on_demand_bytes_per_dollar = 2**40 / 5
min_flat_rate_bytes_per_second_per_slot = (
    on_demand_bytes_per_dollar * flat_rate_dollars_per_year_per_slot / seconds_per_year
)

# translate min_flat_rate_bytes_per_second_per_slot to slot_millis_per_byte
slot_millis_per_second_per_slot = 1000
min_flat_rate_slot_millis_per_byte = (
    slot_millis_per_second_per_slot / min_flat_rate_bytes_per_second_per_slot
)

parser = ArgumentParser()
standard_args.add_argument(
    parser,
    "-s",
    "--slots",
    default=1000,
    type=int,
    help="Number of reserved slots currently available",
)
standard_args.add_argument(
    parser,
    "-S",
    "--state-table",
    "--state_table",
    default="moz-fx-data-shredder.shredder_state.shredder_state",
    help="Table used to store shredder state for script/shredder_delete",
)
standard_args.add_argument(
    parser,
    "-d",
    "--days",
    default=28,
    type=int,
    help="Number of days shredder has to run",
)


def mean(iterable):
    """Calculate the mean of an iterable of optional values.

    Each item in iterable may be a value with an implicit weight of 1 or a
    tuple of value and weight. When value is None exclude it from the mean.
    """
    total_value = total_weight = 0
    for item in iterable:
        if isinstance(item, tuple):
            value, weight = item
        else:
            value, weight = item, 1
        if value is None:
            continue
        total_value += value * weight
        total_weight += weight
    if total_weight != 0:
        return total_value / total_weight


def get_bytes_per_second(jobs, flat_rate_slots, table=None, verbose=False):
    """Calculcate the speed of jobs matching table in bytes per second."""
    if table is not None:
        table_id = sql_table_id(table)
        jobs = [job for job in jobs if sql_table_id(job) == table_id]
    if jobs:
        total_bytes_processed, slot_millis = map(
            sum, zip(*((job.total_bytes_processed, job.slot_millis) for job in jobs))
        )
        if total_bytes_processed == 0 or slot_millis == 0:
            return None
        slot_millis_per_byte = slot_millis / total_bytes_processed
        bytes_per_second = total_bytes_processed / (
            slot_millis / slot_millis_per_second_per_slot / flat_rate_slots
        )
        if verbose:
            print(
                "shredder is processing "
                f"{table.table_id + ' at ' if table else ''}"
                f"{bytes_per_second*60/2**30:.3f} GiB/min "
                f"using {flat_rate_slots} slots"
            )
            # report cost vs on-demand
            efficiency = slot_millis_per_byte / min_flat_rate_slot_millis_per_byte
            tense = "cheaper" if efficiency <= 1 else "more expensive"
            efficiency_percent = abs(100 - efficiency * 100)
            print(
                f"processing speed{' for ' + table.table_id if table else ''} is "
                f"{efficiency_percent:.0f}% {tense} than on-demand at 100% utilization"
            )
        return bytes_per_second
    return None


def main():
    """Report estimated cost to run shredder."""
    args = parser.parse_args()
    flat_rate_slots = args.slots
    runs_per_year = 365 / args.days

    # translate days per run to seconds per run
    seconds_per_run = args.days * seconds_per_day

    # determine how fast shredder is running
    warnings.filterwarnings("ignore", module="google.auth._default")
    client = bigquery.Client()
    jobs = list(client.query(JOBS_QUERY).result())

    table_ids = {sql_table_id(target) for target in DELETE_TARGETS}
    tables = [
        table
        for table in client.query(TABLES_QUERY).result()
        if sql_table_id(table) in table_ids
    ]
    flat_rate_tables = [table for table in tables if table.table_id != "main_v4"]
    on_demand_tables = [table for table in tables if table.table_id == "main_v4"]

    # estimate how much it costs to process flat-rate tables
    flat_rate_speeds_and_bytes = [
        (get_bytes_per_second(jobs, flat_rate_slots, table), table.num_bytes)
        for table in flat_rate_tables
    ]
    # mean weighted by table.num_bytes for tables with jobs
    mean_bytes_per_second = mean(flat_rate_speeds_and_bytes)
    # calculate using mean_bytes_per_second for tables with no jobs
    slots_needed = sum(
        num_bytes
        / seconds_per_run
        / (table_bytes_per_second or mean_bytes_per_second)
        * flat_rate_slots
        for table_bytes_per_second, num_bytes in flat_rate_speeds_and_bytes
    )
    # slots are reserved in increments of 500
    slots_reserved = ceil(slots_needed / 500) * 500
    flat_rate_cost = slots_reserved * flat_rate_dollars_per_year_per_slot
    flat_rate_num_bytes = sum(table.num_bytes for table in flat_rate_tables)

    # estimate how much it costs to process on-demand queries
    on_demand_num_bytes = sum(table.num_bytes for table in on_demand_tables)
    on_demand_cost = on_demand_num_bytes * runs_per_year / on_demand_bytes_per_dollar

    query_cost = flat_rate_cost + on_demand_cost

    # estimate how much it increases costs to store data
    total_num_bytes = on_demand_num_bytes + flat_rate_num_bytes
    added_storage_cost_per_gigabyte_per_year = 0.01 * 12  # $0.01 per GiB
    bytes_per_gigabyte = 2**30
    storage_cost = (
        total_num_bytes * added_storage_cost_per_gigabyte_per_year / bytes_per_gigabyte
    )

    # estimate how much cost will grow due to retention policies
    # days since the start of stable tables
    today = datetime.utcnow().date()
    yesterday = today - timedelta(days=1)
    start_of_stable = datetime(2018, 10, 30).date()
    days_in_tables = (today - start_of_stable).days
    # normalize 25 month retention to 2 years 1 month with 28 day "months"
    days_in_retention = 28 * (13 * 2 + 1)
    # estimate how much data will grow before retention kicks in
    data_growth = 0
    if days_in_tables < days_in_retention:
        data_growth = days_in_retention / days_in_tables - 1
    # translate data_growth to cost
    on_demand_cost_growth = on_demand_cost * data_growth
    storage_cost_growth = storage_cost * data_growth
    total_cost_growth = on_demand_cost_growth + storage_cost_growth
    slots_reserved_growth = (
        ceil(slots_needed * (1 + data_growth) / 500) * 500 - slots_reserved
    )
    flat_rate_growth_msg = (
        "$0 is is flat-rate query cost because existing "
        "rounded up query capacity should suffice."
    )
    if slots_reserved_growth > 0:
        flat_rate_cost_growth = (
            slots_reserved_growth * flat_rate_dollars_per_year_per_slot
        )
        flat_rate_growth_msg = (
            f"${flat_rate_cost_growth:,.0f} is flat-rate query cost for "
            f"{slots_reserved_growth} additional reserved slots"
        )
        total_cost_growth += flat_rate_cost_growth

    total_cost = query_cost + storage_cost + total_cost_growth

    print(
        dedent(
            f"""
            Cost
            ===

            As of {today}, the cost of this project is estimated to be
            ${total_cost:,.0f}/year. Of this, ${query_cost:,.0f} is direct cost of
            scanning and deleting data, ${storage_cost:,.0f} is increased storage cost
            due to active (vs. long-term) storage pricing, and
            ${total_cost_growth:,.0f} is predicted cost increases as more data is
            accumulated and stored.

            On-demand queries
            ---

            telemetry_stable.main_v4 contains {on_demand_num_bytes/2**50:.3f} PiB of
            data, which is scanned by queries using on-demand pricing at $5 per TiB¹
            every {args.days} days, costing ${on_demand_cost:,.0f}/year.

            Flat-rate queries
            ---

            The rest of the tables impacted contain {flat_rate_num_bytes/2**50:.3f} PiB
            of data, which is scanned by queries using flat-rate pricing at $8,500 per
            500 reserved slots¹, and requires approximately {slots_needed:.0f} slots in
            order to complete every {args.days} days, rounded up to {slots_reserved}
            slots, costing ${flat_rate_cost:,.0f}/year.

            Active storage
            ---

            Combined all of the tables contain {total_num_bytes/2**50:.3f} PiB of data,
            which is stored using active storage pricing at $0.02 per GiB per month¹
            and would otherwise be stored using long-term storage pricing at $0.01 per
            GiB per month¹, thus increasing storage cost by ${storage_cost:,.0f}/year.

            Data volume growth
            ---

            The above costs total ${query_cost+storage_cost:,.0f}/year, but that does
            not account for table sizes increasing over time. If all tables had a
            retention policy of 25 months, as of {today} held data from
            {start_of_stable} through {yesterday}, and had a stable daily data volume,
            then total data volume would increase by {100*data_growth:.0f}% before any
            data were dropped due to retention policies.

            Assuming a linear impact on queries, this would increase cost by
            ${total_cost_growth:,.0f}/year. Of this, ${storage_cost_growth:,.0f} is
            storage cost, ${on_demand_cost_growth:,.0f} is on-demand query cost, and
            {flat_rate_growth_msg}.


            ¹ https://cloud.google.com/bigquery/pricing#pricing_summary (this doc
            currently uses the annual $8500 per 500 slots flat-rate query pricing
            rather than the monthly $10000 per 500 slots)
            """
        )
    )


if __name__ == "__main__":
    main()



"""Generate active users aggregates per app."""

import itertools
from collections import ChainMap
from dataclasses import asdict, dataclass, field
from enum import Enum
from os import path
from pathlib import Path
from typing import Any

import click
from jinja2 import Environment, FileSystemLoader

from bigquery_etl.cli.utils import use_cloud_function_option
from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.util.common import write_sql

GENERATOR_ROOT = Path(path.dirname(__file__))

HEADER = f"-- Query generated via `{GENERATOR_ROOT.name}` SQL generator."
VERSION = "v1"
TEMPLATES = (
    ("CLIENT", "active_users.view.sql"),
    ("CLIENT", "retention_clients.view.sql"),
    ("AGGREGATE", "retention.query.sql"),
    ("AGGREGATE", "retention.view.sql"),
    ("CLIENT", "engagement_clients.view.sql"),
    ("AGGREGATE", "engagement.query.sql"),
    ("AGGREGATE", "engagement.view.sql"),
    ("CLIENT", "attribution_clients.view.sql"),
    ("CLIENT", "attribution_clients.query.sql"),
    ("CLIENT", "new_profile_clients.view.sql"),
    ("AGGREGATE", "new_profiles.view.sql"),
    ("AGGREGATE", "new_profiles.query.sql"),
)
BIGEYE_COLLECTION = "Operational Checks"
BIGEYE_NOTIFICATION_SLACK_CHANNEL = "#de-bigeye-triage"


class AttributionPings(Enum):
    """An enumerator containing a list of pings that could be the source of attribution information."""

    metrics: str = "metrics"
    first_session: str = "first_session"
    baseline: str = "baseline"


@dataclass
class AttributionFieldGroup:
    """An object to encapsulate a logical grouping of attribution fields together."""

    name: str
    source_pings: list[AttributionPings]
    fields: list[dict[str, Any]]


class AttributionFields:
    """Defines all possible AttributionFieldGroups."""

    install_source = AttributionFieldGroup(
        name="install_source",
        source_pings=[AttributionPings.metrics],
        fields=[
            {
                "name": "install_source",
                "type": "STRING",
                "description": "The source of a profile installation.",
            },
        ],
    )
    adjust = AttributionFieldGroup(
        name="adjust",
        source_pings=[AttributionPings.metrics, AttributionPings.first_session],
        fields=[
            {
                "name": "adjust_ad_group",
                "type": "STRING",
                "description": "Adjust Ad Group the profile is attributed to.",
            },
            {
                "name": "adjust_campaign",
                "type": "STRING",
                "description": "Adjust Campaign the profile is attributed to.",
            },
            {
                "name": "adjust_creative",
                "type": "STRING",
                "description": "Adjust Creative the profile is attributed to.",
            },
            {
                "name": "adjust_network",
                "type": "STRING",
                "description": "Adjust Network the profile is attributed to.",
            },
            {
                "name": "adjust_attribution_timestamp",
                "type": "TIMESTAMP",
                "description": "Timestamp corresponding to the ping that contained the adjust attribution.",
                "client_only": True,
            },
        ],
    )
    play_store = AttributionFieldGroup(
        name="play_store",
        source_pings=[AttributionPings.first_session],
        fields=[
            {
                "name": "play_store_attribution_campaign",
                "type": "STRING",
                "description": "Play store campaign the profile is attributed to.",
            },
            {
                "name": "play_store_attribution_medium",
                "type": "STRING",
                "description": "Play store Medium the profile is attributed to.",
            },
            {
                "name": "play_store_attribution_source",
                "type": "STRING",
                "description": "Play store source the profile is attributed to.",
            },
            {
                "name": "play_store_attribution_timestamp",
                "type": "TIMESTAMP",
                "description": "Timestamp corresponding to the ping that contained the play_store attribution.",
                "client_only": True,
            },
            {
                "name": "play_store_attribution_content",
                "type": "STRING",
                "description": "",
                "client_only": True,
            },
            {
                "name": "play_store_attribution_term",
                "type": "STRING",
                "description": "",
                "client_only": True,
            },
            {
                "name": "play_store_attribution_install_referrer_response",
                "type": "STRING",
                "description": "Play store source the profile is attributed to.",
                "client_only": True,
            },
        ],
    )
    meta = AttributionFieldGroup(
        name="meta",
        source_pings=[AttributionPings.first_session],
        fields=[
            {
                "name": "meta_attribution_app",
                "type": "STRING",
                "description": "Facebook app linked to paid marketing.",
            },
            {
                "name": "meta_attribution_timestamp",
                "type": "TIMESTAMP",
                "description": "Timestamp corresponding to the ping that contained the meta attribution.",
                "client_only": True,
            },
        ],
    )
    is_suspicious_device_client = AttributionFieldGroup(
        name="is_suspicious_device_client",
        source_pings=[],
        fields=[
            {
                "name": "is_suspicious_device_client",
                "type": "BOOLEAN",
                "description": "Flag to identify suspicious device users, see bug-1846554 for more info.",
            },
        ],
    )
    distribution_id = AttributionFieldGroup(
        name="distribution_id",
        source_pings=[
            AttributionPings.first_session,
            AttributionPings.baseline,
            AttributionPings.metrics,
        ],
        fields=[
            {
                "name": "distribution_id",
                "type": "STRING",
                "description": "A string containing the distribution identifier.",
            },
        ],
    )
    empty = AttributionFieldGroup(
        name="empty",
        source_pings=[],
        fields=[],
    )


@dataclass
class Product:
    """Encapsulation of what we expect a 'Product' to look like in this generator."""

    friendly_name: str
    is_mobile_kpi: bool = False
    attribution_groups: list[AttributionFieldGroup] = field(
        default_factory=list[AttributionFields.empty]  # type: ignore[valid-type]
    )

    def get_product_attribution_fields(self) -> dict[str, dict[str, Any]]:
        """Merge all AttributionFieldGroups assigned to a product into a single map containing all attribution fields set for that product."""
        return {
            field["name"]: field
            for field in list(
                itertools.chain.from_iterable(
                    [
                        attribution_group.fields
                        for attribution_group in self.attribution_groups
                    ]
                )
            )
        }

    def get_attribution_pings(self) -> list[str]:
        """Merge all source_pings of AttributionFieldGroups assigned to \
        a product into a single list of all pings containing attribution data for that specific product."""
        return list(
            set(
                [
                    ping.value
                    for attribution_group in self.attribution_groups
                    for ping in attribution_group.source_pings
                ]
            )
        )

    def get_attribution_group_names(self) -> list[str]:
        """Return a list containing AttributionGroup names only."""
        return [attribution_group.name for attribution_group in self.attribution_groups]


class MobileProducts(Enum):
    """Enumeration with browser names and equivalent dataset names."""

    fenix = Product(
        friendly_name="Fenix",
        is_mobile_kpi=True,
        attribution_groups=[
            AttributionFields.play_store,
            AttributionFields.meta,
            AttributionFields.install_source,
            AttributionFields.adjust,
            AttributionFields.distribution_id,
        ],
    )
    focus_android = Product(
        friendly_name="Focus Android",
        is_mobile_kpi=True,
    )
    klar_android = Product(
        friendly_name="Klar Android",
    )
    firefox_ios = Product(
        friendly_name="Firefox iOS",
        is_mobile_kpi=True,
        attribution_groups=[
            AttributionFields.is_suspicious_device_client,
            AttributionFields.adjust,
        ],
    )
    focus_ios = Product(
        friendly_name="Focus iOS",
        is_mobile_kpi=True,
    )
    klar_ios = Product(
        friendly_name="Klar iOS",
    )


@click.command()
@click.option(
    "--output-dir",
    "--output_dir",
    help="Output directory generated SQL is written to",
    type=click.Path(file_okay=False),
    default="sql",
)
@click.option(
    "--target-project",
    "--target_project",
    help="Google Cloud project ID",
    default="moz-fx-data-shared-prod",
)
@use_cloud_function_option
def generate(target_project, output_dir, use_cloud_function):
    """Generate per-app queries, views and metadata for active users and search counts aggregates.

    The parent folders will be created if not existing and existing files will be overwritten.
    """
    env = Environment(loader=FileSystemLoader(str(GENERATOR_ROOT / "templates")))
    output_dir = Path(output_dir) / target_project

    default_template_args = {
        "header": HEADER,
        "version": VERSION,
        "project_id": target_project,
        "bigeye_collection": BIGEYE_COLLECTION,
        "bigeye_notification_slack_channel": BIGEYE_NOTIFICATION_SLACK_CHANNEL,
    }

    query_support_configs = (
        "metadata.yaml",
        "schema.yaml",
        "bigconfig.yml",
    )

    all_possible_attribution_fields = dict(
        ChainMap(
            *[
                product.value.get_product_attribution_fields()
                for product in MobileProducts
            ]
        )
    )

    for template_grain, template in TEMPLATES:
        for product in MobileProducts:

            target_name, target_filename, target_extension = template.split(".")

            target_dataset = (
                product.name + "_derived"
                if target_filename == "query"
                else product.name
            )

            table_id = f"{target_project}.{target_dataset}.{target_name}"

            full_table_id = (
                table_id + f"_{VERSION}" if target_filename == "query" else table_id
            )

            product_args = {
                "dataset": product.name,
                "target_name": target_name,
                "app_name": product.name,
                "name": target_name,
                "product_attribution_groups": product.value.attribution_groups,
                "product_attribution_group_names": product.value.get_attribution_group_names(),
                "product_attribution_group_pings": product.value.get_attribution_pings(),
                "product_attribution_fields": product.value.get_product_attribution_fields(),
            }

            sql_template = env.get_template(template)
            rendered_sql = reformat(
                sql_template.render(
                    **default_template_args,
                    **asdict(product.value),
                    **product_args,
                )
            )

            write_sql(
                output_dir=output_dir,
                full_table_id=full_table_id,
                basename=f"{target_filename}.{target_extension}",
                sql=rendered_sql,
                skip_existing=False,
            )

            # we only want to copy files defined in query_support_configs for query files.
            if target_filename != "query":
                continue

            for query_support_config in query_support_configs:
                support_config_template = env.get_template(
                    f"{target_name}.{query_support_config}"
                )
                support_config_rendered = support_config_template.render(
                    **default_template_args,
                    **asdict(product.value),
                    **product_args,
                    format=False,
                )

                write_sql(
                    output_dir=output_dir,
                    full_table_id=full_table_id,
                    basename=query_support_config,
                    sql=(
                        reformat(support_config_rendered)
                        if query_support_config.endswith(".sql")
                        else support_config_rendered
                    ),
                    skip_existing=False,
                )

        # we only want to generate a union view inside telemetry for views
        if target_filename != "view":
            continue

        # For now skipping attribution_clients union. Will add in the future.
        if target_name == "attribution_clients":
            continue

        target_dataset = "telemetry"

        union_target_name = f"mobile_{target_name}"

        union_sql_template = env.get_template("union.view.sql")
        union_sql_rendered = union_sql_template.render(
            **default_template_args,
            dataset=target_dataset,
            name=target_name,
            target_name=union_target_name,
            target_filename=target_filename,
            template_grain=template_grain,
            format=False,
            products=[
                {
                    "name": product.name,
                    "all_possible_attribution_fields": (
                        [
                            {
                                "exists": field_name
                                in product.value.get_product_attribution_fields().keys(),
                                "name": field_name,
                                "type": field_properties["type"],
                                "client_only": field_properties.get(
                                    "client_only", False
                                ),
                            }
                            for field_name, field_properties in all_possible_attribution_fields.items()
                        ]
                        if not template.startswith("active_users")
                        else []
                    ),
                }
                for product in MobileProducts
            ],
        )

        write_sql(
            output_dir=output_dir,
            full_table_id=f"{target_project}.{target_dataset}.{union_target_name}",
            basename=f"{target_filename}.{target_extension}",
            sql=(reformat(union_sql_rendered)),
            skip_existing=False,
        )



# Import most recent major versions from: https://product-details.mozilla.org/1.0/firefox_versions.json

from argparse import ArgumentParser
from google.cloud import bigquery
import json
import requests
from pathlib import Path

URL = "https://product-details.mozilla.org/1.0/firefox_versions.json"
SCHEMA_FILE = Path(__file__).parent / "schema.yaml"

parser = ArgumentParser(description=__doc__)
parser.add_argument("--project", default="moz-fx-data-shared-prod")
parser.add_argument("--destination_dataset", default="telemetry_derived")
parser.add_argument("--destination_table", default="latest_versions_v2")


def main():
    args = parser.parse_args()
    client = bigquery.Client(args.project)
    destination_table = (
        f"{args.project}.{args.destination_dataset}.{args.destination_table}"
    )

    versions_json = json.loads(requests.get(URL).text)

    query = f"""
    SELECT 
      'release' AS channel,
      mozfun.norm.extract_version('{versions_json["LATEST_FIREFOX_VERSION"]}', 'major') AS latest_version
    UNION ALL
    SELECT
      'beta' AS channel,
      mozfun.norm.extract_version('{versions_json["FIREFOX_DEVEDITION"]}', 'major') AS latest_version
    UNION ALL
    SELECT
      'nightly' AS channel,
      mozfun.norm.extract_version('{versions_json["FIREFOX_NIGHTLY"]}', 'major') AS latest_version
    UNION ALL
    SELECT
      'esr' AS channel,
      mozfun.norm.extract_version('{versions_json["FIREFOX_ESR"]}', 'major') AS latest_version
    """

    job_config = bigquery.QueryJobConfig(
        destination=destination_table, write_disposition="WRITE_APPEND"
    )
    job_config.write_disposition = bigquery.job.WriteDisposition.WRITE_TRUNCATE
    client.query(query, job_config=job_config).result()


if __name__ == "__main__":
    main()



"""Generate docs for bqetl commands."""

import os
from pathlib import Path

import click
from jinja2 import Environment, FileSystemLoader

from bigquery_etl.cli.backfill import backfill
from bigquery_etl.cli.dag import dag
from bigquery_etl.cli.dryrun import dryrun
from bigquery_etl.cli.format import format
from bigquery_etl.cli.query import query
from bigquery_etl.cli.routine import mozfun, routine
from bigquery_etl.dependency import dependency

# commands to document
COMMANDS = {
    "query": query,
    "dag": dag,
    "dependency": dependency,
    "dryrun": dryrun,
    "format": format,
    "routine": routine,
    "mozfun": mozfun,
    "backfill": backfill,
}
FILE_PATH = Path(os.path.dirname(__file__))
BASE_DIR = FILE_PATH.parent.parent
TEMPLATE = FILE_PATH / "templates" / "commands.md"


def extract_description_from_help_text(help_text):
    """Return the description from the command help text."""
    return help_text.split("\n\n")[0].strip()


def extract_examples_from_help_text(help_text):
    """Return the examples from the command help text."""
    examples = None
    help_text = help_text.split("\n\n")

    if len(help_text) > 1:
        examples = "\n\n".join(help_text[1:])
        examples = examples.replace("Examples:\n\n", "")
        # Examples have \b markers to preserve formatting for the following block
        # when rendered using Click:
        # https://click.palletsprojects.com/en/7.x/documentation/#preventing-rewrapping
        # Removing it, otherwise it shows up as newline in rendered Markdown
        examples = examples.replace("\b", "")
        examples = examples.strip()

    return examples


def extract_options_from_command(cmd):
    """Extract options with descriptions from click command."""
    return [
        {"name": option.name, "description": option.help}
        for option in cmd.params
        if isinstance(option, click.Option)
    ]


def extract_arguments_from_command(cmd):
    """Extract options with descriptions from click command."""
    return [{"name": arg.name} for arg in cmd.params if isinstance(arg, click.Argument)]


def parse_commands(cmd, path):
    """Parse click commands and store in dict."""
    commands = []
    # full command path; reflect click group nesting
    path = path + " " + cmd.name
    if isinstance(cmd, click.Group):
        commands = [parse_commands(c, path) for _, c in cmd.commands.items()]

    return {
        "name": cmd.name,
        "commands": commands,
        "examples": extract_examples_from_help_text(cmd.help),
        "description": extract_description_from_help_text(cmd.help),
        "options": extract_options_from_command(cmd),
        "arguments": extract_arguments_from_command(cmd),
        "path": path,
    }


def generate_bqetl_docs(out_file):
    """Generate documentation for bqetl CLI commands."""
    print("Generate bqetl command docs.")
    command_groups = [parse_commands(group, "") for _, group in COMMANDS.items()]

    # render md docs
    file_loader = FileSystemLoader(TEMPLATE.parent)
    env = Environment(loader=file_loader)
    template = env.get_template(TEMPLATE.name)
    output = template.render(command_groups=command_groups)

    # append to bqetl docs page
    with open(out_file, "a") as out:
        out.write(output)



"""Machinery for exporting query results as JSON to Cloud storage."""

import datetime
import json
import logging
import random
import re
import string
import sys
from argparse import ArgumentParser

import smart_open
from google.cloud import storage  # type: ignore
from google.cloud import bigquery

from bigquery_etl.config import ConfigLoader
from bigquery_etl.metadata.parse_metadata import Metadata
from bigquery_etl.metadata.validate_metadata import validate_public_data

SUBMISSION_DATE_RE = re.compile(r"^submission_date:DATE:(\d\d\d\d-\d\d-\d\d)$")
QUERY_FILE_RE = re.compile(r"^.*/([a-zA-Z0-9_]+)/([a-zA-Z0-9_]+)_(v[0-9]+)/query\.sql$")
MAX_JSON_SIZE = 1 * 1024 * 1024 * 1024  # 1 GB as max. size of exported JSON files
# maximum number of JSON output files, output file names are only up to 12 characters
MAX_FILE_COUNT = 10_000
# exported file name format: 000000000000.json, 000000000001.json, ...
MAX_JSON_NAME_LENGTH = 12

logging.basicConfig(
    level=logging.DEBUG, format="%(asctime)s: %(levelname)s: %(message)s"
)


class JsonPublisher:
    """Publishes query results as JSON."""

    def __init__(
        self,
        client,
        storage_client,
        project_id,
        query_file,
        api_version,
        target_bucket,
        parameter=None,
        gcs_path="",
    ):
        """Init JsonPublisher."""
        self.project_id = project_id
        self.query_file = query_file
        self.api_version = api_version
        self.target_bucket = target_bucket
        self.gcs_path = gcs_path
        self.parameter = parameter
        self.client = client
        self.storage_client = storage_client
        self.temp_table = None
        self.date = None
        self.stage_gcs_path = (
            self.gcs_path
            + "stage/json/"
            + "".join(random.choice(string.ascii_lowercase) for i in range(12))
            + "/"
        )

        self.metadata = Metadata.of_query_file(self.query_file)

        # only for incremental exports files are written into separate directories
        # for each date, ignore date parameters for non-incremental exports
        if self.metadata.is_incremental_export() and self.parameter:
            for p in self.parameter:
                date_search = re.search(SUBMISSION_DATE_RE, p)

                if date_search:
                    self.date = date_search.group(1)

        query_file_re = re.search(QUERY_FILE_RE, self.query_file)
        if query_file_re:
            self.dataset = query_file_re.group(1)
            self.table = query_file_re.group(2)
            self.version = query_file_re.group(3)
        else:
            logging.error("Invalid file naming format: %s", self.query_file)
            sys.exit(1)

    def _clear_stage_directory(self):
        """Delete files in stage directory."""
        tmp_blobs = self.storage_client.list_blobs(
            self.target_bucket, prefix=self.stage_gcs_path
        )

        for tmp_blob in tmp_blobs:
            tmp_blob.delete()

    def publish_json(self):
        """Publish query results as JSON to GCP Storage bucket."""
        self.last_updated = datetime.datetime.utcnow()

        try:
            if self.metadata.is_incremental_export():
                if self.date is None:
                    logging.error(
                        "Cannot publish JSON. submission_date missing in parameter."
                    )
                    sys.exit(1)

                # if it is an incremental query, then the query result needs to be
                # written to a temporary table to get exported as JSON
                self._write_results_to_temp_table()
                self._publish_table_as_json(self.temp_table)
            else:
                # for non-incremental queries, the entire destination table is exported
                result_table = f"{self.dataset}.{self.table}_{self.version}"
                self._publish_table_as_json(result_table)

            self._publish_last_updated()
        finally:
            # delete temporary artifacts
            if self.temp_table:
                self.client.delete_table(self.temp_table)

            self._clear_stage_directory()

    def _publish_table_as_json(self, result_table):
        """Export the `result_table` data as JSON to Cloud Storage."""
        prefix = (
            f"{self.gcs_path}api/{self.api_version}/tables/{self.dataset}/"
            f"{self.table}/{self.version}/files/"
        )

        if self.date is not None:
            # if date exists, then query is incremental and newest results are exported
            prefix += f"{self.date}/"

        logging.info(f"""Export JSON for {result_table} to {self.stage_gcs_path}""")

        table_ref = self.client.get_table(result_table)

        job_config = bigquery.ExtractJobConfig()
        job_config.destination_format = "NEWLINE_DELIMITED_JSON"

        # "*" makes sure that files larger than 1GB get split up into JSON files
        # files are written to a stage directory first
        destination_uri = f"gs://{self.target_bucket}/{self.stage_gcs_path}*.ndjson"
        extract_job = self.client.extract_table(
            table_ref, destination_uri, location="US", job_config=job_config
        )
        extract_job.result()

        self._gcp_convert_ndjson_to_json(prefix)

    def _gcp_convert_ndjson_to_json(self, gcs_path):
        """Convert ndjson files on GCP to json files."""
        blobs = self.storage_client.list_blobs(
            self.target_bucket, prefix=self.stage_gcs_path
        )
        bucket = self.storage_client.bucket(self.target_bucket)

        # keeps track of the number of bytes written to the JSON output file
        output_size = 0
        # tracks the current output file name
        output_file_counter = 0
        # track if the first JSON object is currently processed
        first_line = True
        # output file handler
        output_file = None

        for blob in blobs:
            blob_path = f"gs://{self.target_bucket}/{blob.name}"

            # stream from GCS
            with smart_open.open(blob_path) as fin:
                for line in fin:
                    # check if JSON output file reached limit
                    if output_file is None or output_size >= MAX_JSON_SIZE:
                        if output_file is not None:
                            output_file.write("]")
                            output_file.close()

                        if output_file_counter >= MAX_FILE_COUNT:
                            logging.error(
                                "Maximum number of JSON output files reached."
                            )
                            sys.exit(1)

                        tmp_blob_name = "/".join(blob.name.split("/")[:-1])
                        tmp_blob_name += (
                            "/"
                            + str(output_file_counter).zfill(MAX_JSON_NAME_LENGTH)
                            + ".json.tmp.gz"
                        )
                        tmp_blob_path = f"gs://{self.target_bucket}/{tmp_blob_name}"

                        logging.info(f"""Write {blob_path} to {tmp_blob_path}""")

                        output_file = smart_open.open(tmp_blob_path, "w")
                        output_file.write("[")
                        first_line = True
                        output_file_counter += 1
                        output_size = 1

                    # skip the first line, it has no preceding json object
                    if not first_line:
                        output_file.write(",")

                    output_file.write(line.replace("\n", ""))
                    output_size += len(line)
                    first_line = False

        output_file.write("]")
        output_file.close()

        # move all files from stage directory to target directory
        tmp_blobs = self.storage_client.list_blobs(
            self.target_bucket, prefix=self.stage_gcs_path
        )

        for tmp_blob in tmp_blobs:
            # only copy gzipped files to target directory
            if "tmp.gz" in tmp_blob.name:
                # remove .tmp from the final file name
                file_name = tmp_blob.name.split("/")[-1].replace(".tmp.gz", "")

                logging.info(f"""Move {tmp_blob.name} to {gcs_path + file_name}""")

                bucket.rename_blob(tmp_blob, gcs_path + file_name)

                # set Content-Type to json and encoding to gzip
                blob = self.storage_client.get_bucket(self.target_bucket).get_blob(
                    gcs_path + file_name
                )
                blob.content_type = "application/json"
                blob.content_encoding = "gzip"
                blob.patch()

    def _write_results_to_temp_table(self):
        """Write the query results to a temporary table and return the table name."""
        table_date = self.date.replace("-", "") + "".join(
            random.choice(string.ascii_lowercase) for i in range(12)
        )

        self.temp_table = (
            f"{self.project_id}.tmp.{self.table}_{self.version}_{table_date}_temp"
        )

        with open(self.query_file) as query_stream:
            sql = query_stream.read()

            query_parameters = []
            for p in self.parameter:
                [name, parameter_type, value] = p.split(":")
                query_parameters.append(
                    bigquery.ScalarQueryParameter(name, parameter_type, value)
                )

            job_config = bigquery.QueryJobConfig(
                destination=self.temp_table, query_parameters=query_parameters
            )
            query_job = self.client.query(sql, job_config=job_config)
            query_job.result()

    def _publish_last_updated(self):
        """Write the timestamp when file of the dataset were last modified to GCS."""
        last_updated_path = (
            f"{self.gcs_path}api/{self.api_version}/tables/{self.dataset}/"
            f"{self.table}/{self.version}/last_updated"
        )
        output_file = f"gs://{self.target_bucket}/{last_updated_path}"

        logging.info(f"Write last_updated to {output_file}")

        with smart_open.open(output_file, "w") as fout:
            last_updated = self.last_updated.strftime("%Y-%m-%d %H:%M:%S")
            fout.write(json.dumps(last_updated))

        # set Content-Type to json so that timestamp is displayed in the browser
        blob = self.storage_client.get_bucket(self.target_bucket).get_blob(
            last_updated_path
        )
        blob.content_type = "application/json"
        blob.patch()


parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--target-bucket",
    "--target_bucket",
    default=ConfigLoader.get(
        "public_data", "bucket", fallback="mozilla-public-data-http"
    ),
    help="GCP bucket JSON data is exported to",
)
parser.add_argument(
    "--public_project_id",
    "--public-project-id",
    default="mozilla-public-data",
    help="Run query in the target project",
)
parser.add_argument(
    "--api_version",
    "--api-version",
    default=ConfigLoader.get("public_data", "api_version", fallback="v1"),
    help="API version data is published under in the storage bucket",
)
parser.add_argument(
    "--parameter", action="append", help="Query parameters, such as submission_date"
)
parser.add_argument(
    "--query-file", "--query_file", help="File path to query to be executed"
)
parser.add_argument(
    "--gcs-path", "--gcs_path", default="", help="GCS path data is exported to"
)


def main():
    """Publish query data as JSON to GCS."""
    args, query_arguments = parser.parse_known_args()

    try:
        metadata = Metadata.of_query_file(args.query_file)
    except FileNotFoundError:
        print("No metadata file for: {}".format(args.query_file))
        return

    # check if the data should be published as JSON
    if not metadata.is_public_json():
        return

    if not validate_public_data(metadata, args.query_file):
        sys.exit(1)

    storage_client = storage.Client()
    client = bigquery.Client(args.public_project_id)

    publisher = JsonPublisher(
        client,
        storage_client,
        args.public_project_id,
        args.query_file,
        args.api_version,
        args.target_bucket,
        args.parameter,
        args.gcs_path,
    )
    publisher.publish_json()


if __name__ == "__main__":
    main()



"""bigquery-etl CLI alchemer command."""

from datetime import date, datetime, timedelta

import rich_click as click

from bigquery_etl.alchemer.survey import get_survey_data, insert_to_bq


@click.group(
    help="""Commands for importing alchemer data.

    Examples:

    \b
    # Import data from alchemer (surveygizmo) surveys into BigQuery.
    # The date range is inclusive of the start and end values.
    $ ./bqetl alchemer backfill --start-date=2021-01-01 \\
        --end-date=2021-02-01 \\
        --survey_id=xxxxxxxxxxx \\
        --api_token=xxxxxxxxxxxxxx \\
        --api_secret=xxxxxxxxxxxxxxx \\
        --destination_table=moz-fx-data-shared-prod.telemetry_derived.survey_gizmo_daily_attitudes
    """
)
def alchemer():
    """Create the CLI group for the alchemer command."""
    pass


@alchemer.command()
@click.option("--start-date", type=click.DateTime(formats=["%Y-%m-%d"]), required=True)
@click.option(
    "--end-date", type=click.DateTime(formats=["%Y-%m-%d"]), default=str(date.today())
)
@click.option("--survey_id", required=True)
@click.option("--api_token", required=True)
@click.option("--api_secret", required=True)
@click.option("--destination_table", required=True)
def backfill(start_date, end_date, survey_id, api_token, api_secret, destination_table):
    """Import data from alchemer (surveygizmo) surveys into BigQuery.

    The date range is inclusive of the start and end values.
    """
    print(
        f"Runing backfill of {survey_id} from {start_date} to {end_date}"
        " into {destination_table}"
    )
    days = (end_date - start_date).days + 1
    start = datetime.utcnow()
    for i in range(days):
        current_date = (start_date + timedelta(i)).isoformat()[:10]
        print(f"Running for {current_date}")
        survey_data = get_survey_data(survey_id, current_date, api_token, api_secret)
        if not survey_data:
            print("No data, skipping insertion...")
            continue
        insert_to_bq(survey_data, destination_table, current_date)
    print(
        f"Processed {days} days in {int((datetime.utcnow()-start).total_seconds())} seconds"
    )



"""Funnel generation."""

import os
import re
from pathlib import Path

import cattrs
import click
import toml
from jinja2 import Environment, FileSystemLoader

from bigquery_etl.cli.utils import use_cloud_function_option
from bigquery_etl.format_sql.formatter import reformat
from bigquery_etl.util.common import write_sql
from sql_generators.funnels.config import FunnelConfig

FILE_PATH = Path(os.path.dirname(__file__))
TEMPLATES_PATH = FILE_PATH / "templates"


def bq_normalize_name(name: str) -> str:
    return re.sub(r"[^a-zA-Z0-9_]", "_", name)


def generate_funnels(target_project, path, output_dir):
    output_dir = Path(output_dir) / target_project
    path = Path(path)
    converter = cattrs.BaseConverter()

    for config_file in path.glob("*.toml"):
        config = converter.structure(toml.load(config_file), FunnelConfig)
        table_name = bq_normalize_name(config_file.stem)

        env = Environment(loader=FileSystemLoader(TEMPLATES_PATH))
        sql_template = env.get_template("funnel.sql")

        funnel_sql = reformat(
            sql_template.render(
                funnels=config.funnels,
                steps=config.steps,
                data_sources=config.data_sources,
                dimensions=config.dimensions,
                config=config,
            )
        )
        write_sql(
            output_dir=output_dir,
            full_table_id=f"{target_project}.{config.destination_dataset}.{table_name}_v{config.version}",
            basename="query.sql",
            sql=funnel_sql,
            skip_existing=False,
        )

        metadata_template = env.get_template("metadata.yaml")
        rendered_metadata = metadata_template.render(
            {
                "funnel_name": table_name.replace("_", " ").title(),
                "owners": config.owners,
            }
        )
        (
            output_dir
            / config.destination_dataset
            / f"{table_name}_v{config.version}"
            / "metadata.yaml"
        ).write_text(rendered_metadata + "\n")


@click.command("generate")
@click.option(
    "--target-project",
    "--target_project",
    help="Which project the queries should be written to.",
    default="moz-fx-data-shared-prod",
)
@click.option(
    "--path",
    help="Where query directories will be searched for.",
    default="sql_generators/funnels/configs",
    required=False,
    type=click.Path(file_okay=False),
)
@click.option(
    "--output-dir",
    "--output_dir",
    help="The location to write to. Defaults to sql/.",
    default=Path("sql"),
    type=click.Path(file_okay=False),
)
@use_cloud_function_option
def generate(target_project, path, output_dir, use_cloud_function):
    """Generate the funnel queries."""
    output_dir = Path(output_dir)
    generate_funnels(target_project, path, output_dir)



#!/usr/bin/env python3

"""
Generate queries combining metrics for each dimension in the exported app store data.

It is expected that exported tables follow a naming convention
of {metric}_by_{dimension} or {metric}_total.
"""

import os
import sys
from argparse import ArgumentParser
from collections import defaultdict

from google.cloud import bigquery

sys.path.append(
    os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
)
from bigquery_etl.format_sql.formatter import reformat  # noqa E402

SQL_DIR = "sql/moz-fx-data-marketing-prod/"

QUERY_TEMPLATE = """
SELECT
  * EXCEPT ({excepted_fields}),
  {additional_fields}
FROM
  {first_table}
{joined_tables}
WHERE
  {filter}
"""

JOIN_TEMPLATE = """
FULL JOIN
  {table}
  USING ({fields})
"""


def parse_args():
    """Parse command line arguments."""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument(
        "--project",
        default="moz-fx-data-marketing-prod",
        help="Project containing exported app store data",
    )
    parser.add_argument(
        "--source-dataset",
        default="apple_app_store_exported",
        help="Dataset containing exported app store data",
    )
    parser.add_argument(
        "--destination-dataset",
        default="apple_app_store",
        help="Dataset to write queries to",
    )
    parser.add_argument("--sql-dir", default=SQL_DIR)
    parser.add_argument("--dry-run", action="store_true")
    parser.add_argument("--create-table", action="store_true")
    parser.add_argument(
        "--backfill",
        action="store_true",
        help="If --create-table is set, backfill will populate"
        " tables with data from exported tables",
    )

    return parser.parse_args()


def main(project, source_dataset, destination_dataset, create_table, backfill, dryrun):
    """Generate queries and optionally create the tables in BigQuery."""
    client = bigquery.Client(project=project)

    exported_tables = [
        table.table_id
        for table in client.list_tables(source_dataset)
        if table.table_type == "TABLE"
    ]

    tables_by_dimension = defaultdict(list)
    opt_in_metrics = set()

    # group table names by the dimension it is grouped by
    for table_name in exported_tables:
        if table_name.endswith("_total"):
            dimension = None
        else:
            metric, dimension = table_name.split("_by_")
            if dimension.startswith("opt_in"):
                opt_in_metrics.add(metric)
                dimension = dimension.replace("opt_in_", "")

        tables_by_dimension[dimension].append(table_name)

    for dimension, table_names in tables_by_dimension.items():
        qualified_table_names = [
            f"`{project}.{source_dataset}.{table_name}`" for table_name in table_names
        ]

        if dimension is not None:
            fields = f"date, app_name, {dimension}"
            table_name = f"metrics_by_{dimension}"
            metrics = [table_name.split("_by_")[0] for table_name in table_names]
        else:
            fields = "date, app_name"
            table_name = "metrics_total"
            metrics = [table_name.split("_total")[0] for table_name in table_names]

        join_clauses = [
            JOIN_TEMPLATE.format(table=table_name, fields=fields)
            for table_name in qualified_table_names[1:]
        ]

        # add _opt_in to opt-in metrics
        fields_to_add_opt_in = [
            metric for metric in metrics if metric in opt_in_metrics
        ]
        excepted_fields = ",".join(fields_to_add_opt_in)
        additional_fields = [
            f"{name} AS {name}_opt_in"
            for name in fields_to_add_opt_in
            if name != "rate"
        ]

        # rename rate column to opt_in_rate and
        if "rate" in metrics:
            additional_fields.append("rate AS opt_in_rate")

        query_text = QUERY_TEMPLATE.format(
            excepted_fields=excepted_fields,
            additional_fields=", ".join(additional_fields),
            first_table=qualified_table_names[0],
            joined_tables="\n".join(join_clauses),
            filter="date=@submission_date",
        )
        query_path = os.path.join(SQL_DIR, destination_dataset, table_name, "query.sql")

        if not os.path.exists(os.path.dirname(query_path)):
            os.makedirs(os.path.dirname(query_path))

        with open(query_path, "w") as f:
            print(f"Writing {query_path}")
            f.write(reformat(query_text))
            f.write("\n")

        if create_table:
            query_text = QUERY_TEMPLATE.format(
                excepted_fields=excepted_fields,
                additional_fields=", ".join(additional_fields),
                first_table=qualified_table_names[0],
                joined_tables="\n".join(join_clauses),
                filter="TRUE" if backfill else "FALSE",
            )
            schema_update_options = (
                [] if backfill else [bigquery.SchemaUpdateOption.ALLOW_FIELD_ADDITION]
            )
            job_config = bigquery.QueryJobConfig(
                use_legacy_sql=False,
                dry_run=dryrun,
                destination=f"{project}.{destination_dataset}.{table_name}",
                schema_update_options=schema_update_options,
                time_partitioning=bigquery.TimePartitioning(field="date"),
                create_disposition=bigquery.CreateDisposition.CREATE_IF_NEEDED,
                write_disposition=(
                    bigquery.WriteDisposition.WRITE_TRUNCATE
                    if backfill
                    else bigquery.WriteDisposition.WRITE_APPEND
                ),
            )
            print(f"Creating table {table_name}")
            query_job = client.query(query_text, job_config)
            if not dryrun:
                query_job.result()


if __name__ == "__main__":
    args = parse_args()
    main(
        args.project,
        args.source_dataset,
        args.destination_dataset,
        args.create_table,
        args.backfill,
        args.dry_run,
    )



import copy
from uuid import uuid4

import pytest
import requests
from click.testing import CliRunner
from google.cloud import bigquery
from requests import HTTPError

from bigquery_etl.alchemer.survey import (
    construct_data,
    date_plus_one,
    format_responses,
    get_survey_data,
    insert_to_bq,
    main,
    response_schema,
    utc_date_to_eastern_string,
)

# https://apihelp.alchemer.com/help/surveyresponse-returned-fields-v5#getobject
EXAMPLE_RESPONSE = {
    "result_ok": True,
    "total_count": 2,
    "page": 1,
    "total_pages": 1,
    "results_per_page": 50,
    "data": [
        {
            "id": "1",
            "contact_id": "",
            "status": "Complete",
            "is_test_data": "0",
            "date_submitted": "2018-09-27 10:42:26 EDT",
            "session_id": "1538059336_5bacec4869caa2.27680217",
            "language": "English",
            "date_started": "2018-09-27 10:42:16 EDT",
            "link_id": "7473882",
            "url_variables": [],
            "ip_address": "50.232.185.226",
            "referer": "https://app.alchemer.com/distribute/share/id/4599075",
            "user_agent": (
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) "
                "AppleWebKit/537.36 (KHTML, like Gecko) "
                "Chrome/69.0.3497.100 Safari/537.36"
            ),
            "response_time": 10,
            "data_quality": [],
            "longitude": "-105.20369720459",
            "latitude": "40.050701141357",
            "country": "United States",
            "city": "Boulder",
            "region": "CO",
            "postal": "80301",
            "dma": "751",
            "survey_data": {
                "2": {
                    "id": 2,
                    "type": "RADIO",
                    "question": "Will you attend the event?",
                    "section_id": 1,
                    "original_answer": "Yes",
                    "answer": "1",
                    "answer_id": 10001,
                    "shown": True,
                },
                "3": {
                    "id": 3,
                    "type": "TEXTBOX",
                    "question": "How many guests will you bring?",
                    "section_id": 1,
                    "answer": "3",
                    "shown": True,
                },
                "4": {
                    "id": 4,
                    "type": "TEXTBOX",
                    "question": "How many guests are under the age of 18?",
                    "section_id": 1,
                    "answer": "2",
                    "shown": True,
                },
            },
        },
        {
            "id": "2",
            "contact_id": "",
            "status": "Complete",
            "is_test_data": "0",
            "date_submitted": "2018-09-27 10:43:11 EDT",
            "session_id": "1538059381_5bacec751e41f4.51482165",
            "language": "English",
            "date_started": "2018-09-27 10:43:01 EDT",
            "link_id": "7473882",
            "url_variables": {
                "__dbget": {"key": "__dbget", "value": "true", "type": "url"}
            },
            "ip_address": "50.232.185.226",
            "referer": "",
            "user_agent": (
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) "
                "AppleWebKit/537.36 (KHTML, like Gecko) "
                "Chrome/69.0.3497.100 Safari/537.36"
            ),
            "response_time": 10,
            "data_quality": [],
            "longitude": "-105.20369720459",
            "latitude": "40.050701141357",
            "country": "United States",
            "city": "Boulder",
            "region": "CO",
            "postal": "80301",
            "dma": "751",
            "survey_data": {
                "2": {
                    "id": 2,
                    "type": "RADIO",
                    "question": "Will you attend the event?",
                    "section_id": 1,
                    "original_answer": "1",
                    "answer": "1",
                    "answer_id": 10001,
                    "shown": True,
                },
                "3": {
                    "id": 3,
                    "type": "TEXTBOX",
                    "question": "How many guests will you bring?",
                    "section_id": 1,
                    "answer": "2",
                    "shown": True,
                },
                "4": {
                    "id": 4,
                    "type": "TEXTBOX",
                    "question": "How many guests are under the age of 18?",
                    "section_id": 1,
                    "answer": "0",
                    "shown": True,
                },
            },
        },
    ],
}

SUBMISSION_DATE = "2021-01-05"

EXAMPLE_RESPONSE_FORMATTED_0 = {
    "submission_date": SUBMISSION_DATE,
    "id": "1",
    "status": "Complete",
    "session_id": "1538059336_5bacec4869caa2.27680217",
    "response_time": 10,
    "survey_data": [
        {
            "id": 2,
            "type": "RADIO",
            "question": "Will you attend the event?",
            "section_id": 1,
            "original_answer": "Yes",
            "answer": "1",
            "answer_id": 10001,
            "shown": True,
        },
        {
            "id": 3,
            "type": "TEXTBOX",
            "question": "How many guests will you bring?",
            "section_id": 1,
            "answer": "3",
            "shown": True,
        },
        {
            "id": 4,
            "type": "TEXTBOX",
            "question": "How many guests are under the age of 18?",
            "section_id": 1,
            "answer": "2",
            "shown": True,
        },
    ],
}

EXAMPLE_RESPONSE_FORMATTED = [
    EXAMPLE_RESPONSE_FORMATTED_0,
    {
        "submission_date": SUBMISSION_DATE,
        "id": "2",
        "status": "Complete",
        "session_id": "1538059381_5bacec751e41f4.51482165",
        "response_time": 10,
        "survey_data": [
            {
                "id": 2,
                "type": "RADIO",
                "question": "Will you attend the event?",
                "section_id": 1,
                "original_answer": "1",
                "answer": "1",
                "answer_id": 10001,
                "shown": True,
            },
            {
                "id": 3,
                "type": "TEXTBOX",
                "question": "How many guests will you bring?",
                "section_id": 1,
                "answer": "2",
                "shown": True,
            },
            {
                "id": 4,
                "type": "TEXTBOX",
                "question": "How many guests are under the age of 18?",
                "section_id": 1,
                "answer": "0",
                "shown": True,
            },
        ],
    },
]


@pytest.fixture()
def testing_client():
    bq = bigquery.Client()
    yield bq


@pytest.fixture()
def testing_dataset(testing_client):
    bq = testing_client
    dataset_id = f"test_survey_pytest_{str(uuid4())[:8]}"
    bq.delete_dataset(dataset_id, delete_contents=True, not_found_ok=True)
    dataset = bq.create_dataset(dataset_id)
    yield dataset
    bq.delete_dataset(dataset_id, delete_contents=True, not_found_ok=True)


@pytest.fixture()
def testing_table_id(testing_dataset):
    table_ref = testing_dataset.table(f"survey_testing_table_{str(uuid4())[:8]}")
    table_id = f"{table_ref.dataset_id}.{table_ref.table_id}"
    yield table_id


class MockResponse:
    @staticmethod
    def raise_for_status():
        pass

    @staticmethod
    def json():
        return EXAMPLE_RESPONSE

    @property
    def text(self):
        return str(EXAMPLE_RESPONSE)


class MockErrorResponse(MockResponse):
    @staticmethod
    def raise_for_status():
        raise HTTPError()


@pytest.fixture()
def patch_api_requests(monkeypatch):
    # Note: this does not test iterating over multiple pages
    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)


@pytest.fixture()
def patch_api_requests_error(monkeypatch):
    # Note: this does not test iterating over multiple pages
    def mock_get(*args, **kwargs):
        return MockErrorResponse()

    monkeypatch.setattr(requests, "get", mock_get)


def test_utc_date_to_eastern_time():
    # UTC-5 during standard time: https://en.wikipedia.org/wiki/Eastern_Time_Zone
    assert utc_date_to_eastern_string("2021-01-05") == "2021-01-04+19:00:00"


def test_date_plus_one():
    assert date_plus_one("2020-01-05") == "2020-01-06"


def test_format_response():
    assert (
        format_responses(EXAMPLE_RESPONSE["data"][0], SUBMISSION_DATE)
        == EXAMPLE_RESPONSE_FORMATTED_0
    )


def test_format_response_nonnumeric_answer_id():
    base = {
        "submission_date": SUBMISSION_DATE,
        "id": "1",
        "status": "Complete",
        "session_id": "1538059336_5bacec4869caa2.27680217",
        "response_time": 10,
        "survey_data": {
            "1": {
                "answer_id": "10001-other",
            },
            "2": {
                "answer_id": "fadfasdf-other",
            },
        },
    }
    res = format_responses(base, SUBMISSION_DATE)
    assert res["survey_data"][0]["answer_id"] == 10001
    assert not res["survey_data"][1].get("answer_id")


def test_construct_data():
    assert (
        construct_data(EXAMPLE_RESPONSE, SUBMISSION_DATE) == EXAMPLE_RESPONSE_FORMATTED
    )


def test_get_survey_data(patch_api_requests):
    assert (
        get_survey_data("555555", SUBMISSION_DATE, "token", "secret")
        == EXAMPLE_RESPONSE_FORMATTED
    )


def test_get_survey_data_error(patch_api_requests_error):
    """Test that the wrapper correctly reraises the HTTPError."""
    pytest.raises(
        HTTPError, get_survey_data, "555555", SUBMISSION_DATE, "token", "secret"
    )


def test_response_schema():
    # ensure that there aren't any exceptions
    assert response_schema()


@pytest.mark.integration
def test_insert_to_bq(testing_table_id):
    transformed = construct_data(EXAMPLE_RESPONSE, SUBMISSION_DATE)
    insert_to_bq(transformed, testing_table_id, SUBMISSION_DATE)


@pytest.mark.integration
def test_insert_to_bq_options(testing_table_id):
    # Override survey data, but make sure to deep copy to prevent mutating state
    # in other tests.
    # https://apihelp.alchemer.com/help/surveyresponse-per-question-v5#textboxlist
    base = copy.deepcopy(EXAMPLE_RESPONSE["data"][0])
    base["survey_data"] = {
        "37": {
            "id": 37,
            "type": "parent",
            "question": "Textbox List Question Title",
            "section_id": 3,
            "options": {
                "10068": {"id": 10068, "option": "Row 1", "answer": "text list answer"}
            },
            "shown": True,
        },
        "38": {
            "id": 38,
            "type": "parent",
            "question": "Continuous Sum Question Title",
            "section_id": 3,
            "options": {
                "10070": {"id": 10070, "option": "Row 1", "answer": "6"},
                "10071": {"id": 10071, "option": "Row 2", "answer": "7"},
            },
            "shown": True,
        },
    }
    transformed = [format_responses(base, SUBMISSION_DATE)]
    insert_to_bq(transformed, testing_table_id, SUBMISSION_DATE)


@pytest.mark.integration
def test_insert_to_bq_subquestions(testing_table_id):
    # Override survey data. Note that the subquestion object is incompatible.
    # https://apihelp.alchemer.com/help/surveyresponse-per-question-v5#checkboxgrid
    base = copy.deepcopy(EXAMPLE_RESPONSE["data"][0])
    base["survey_data"] = {
        "30": {
            "id": 30,
            "type": "parent",
            "question": "Checkbox Grid Question Title",
            "subquestions": {
                "31": {
                    "10062": {
                        "id": 10062,
                        "type": "CHECKBOX",
                        "parent": 30,
                        "question": "Row 1 : Column 1",
                        "answer": "Column 1",
                        "shown": True,
                    },
                    "10063": {
                        "id": 10063,
                        "type": "CHECKBOX",
                        "parent": 30,
                        "question": "Row 1 : Column 2",
                        "answer": None,
                        "shown": True,
                    },
                },
                "32": {
                    "10062": {
                        "id": 10062,
                        "type": "CHECKBOX",
                        "parent": 30,
                        "question": "Row 2 : Column 1",
                        "answer": None,
                        "shown": True,
                    },
                    "10063": {
                        "id": 10063,
                        "type": "CHECKBOX",
                        "parent": 30,
                        "question": "Row 2 : Column 2",
                        "answer": "Column 2",
                        "shown": True,
                    },
                },
            },
            "section_id": 3,
            "shown": True,
        },
        "83": {
            "id": 83,
            "type": "parent",
            "question": "Custom Table Question Title",
            "subquestions": {
                "10001": {
                    "id": 10001,
                    "type": "RADIO",
                    "question": "Radio Button Column",
                    "section_id": 4,
                    "answer": "Option 1",
                    "answer_id": 10113,
                    "shown": True,
                },
                "10002": {
                    "id": 10002,
                    "type": "RADIO",
                    "question": "Radio Button Column",
                    "section_id": 4,
                    "answer": "Option 2",
                    "answer_id": 10114,
                    "shown": True,
                },
            },
            "section_id": 4,
            "shown": True,
        },
    }
    transformed = [format_responses(base, SUBMISSION_DATE)]
    insert_to_bq(transformed, testing_table_id, SUBMISSION_DATE)


@pytest.mark.integration
def test_cli(patch_api_requests, testing_table_id):
    res = CliRunner().invoke(
        main,
        [
            "--date",
            SUBMISSION_DATE,
            "--survey_id",
            "55555",
            "--api_token",
            "token",
            "--api_secret",
            "secret",
            "--destination_table",
            testing_table_id,
        ],
        catch_exceptions=False,
    )
    assert res.exit_code == 0



"""Query schema."""

import json
import os
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Iterable, List, Optional

import attr
import yaml
from google.api_core.exceptions import NotFound
from google.cloud import bigquery
from google.cloud.bigquery import SchemaField

from .. import dryrun

SCHEMA_FILE = "schema.yaml"


@attr.s(auto_attribs=True)
class Schema:
    """Query schema representation and helpers."""

    schema: Dict[str, Any]
    _type_mapping: Dict[str, str] = {
        "INT64": "INTEGER",
        "BOOL": "BOOLEAN",
        "FLOAT64": "FLOAT",
    }

    @classmethod
    def from_query_file(cls, query_file: Path, *args, **kwargs):
        """Create schema from a query file."""
        if not query_file.is_file() or query_file.suffix != ".sql":
            raise Exception(f"{query_file} is not a valid SQL file.")

        schema = dryrun.DryRun(str(query_file), *args, **kwargs).get_schema()
        return cls(schema)

    @classmethod
    def from_schema_file(cls, schema_file: Path):
        """Create schema from a yaml schema file."""
        if not schema_file.is_file() or schema_file.suffix != ".yaml":
            raise Exception(f"{schema_file} is not a valid YAML schema file.")

        with open(schema_file) as file:
            schema = yaml.load(file, Loader=yaml.FullLoader)
            return cls(schema)

    @classmethod
    def empty(cls):
        """Create an empty schema."""
        return cls({"fields": []})

    @classmethod
    def from_json(cls, json_schema):
        """Create schema from JSON object."""
        return cls(json_schema)

    @classmethod
    def for_table(cls, project, dataset, table, partitioned_by=None, *args, **kwargs):
        """Get the schema for a BigQuery table."""
        query = f"SELECT * FROM `{project}.{dataset}.{table}`"

        if partitioned_by:
            query += f" WHERE DATE(`{partitioned_by}`) = DATE('2020-01-01')"

        try:
            return cls(
                dryrun.DryRun(
                    os.path.join(project, dataset, table, "query.sql"),
                    query,
                    project=project,
                    dataset=dataset,
                    table=table,
                    *args,
                    **kwargs,
                ).get_schema()
            )
        except Exception as e:
            print(f"Cannot get schema for {project}.{dataset}.{table}: {e}")
            return cls({"fields": []})

    def deploy(self, destination_table: str) -> bigquery.Table:
        """Deploy the schema to BigQuery named after destination_table."""
        client = bigquery.Client()
        tmp_schema_file = NamedTemporaryFile()
        self.to_json_file(Path(tmp_schema_file.name))
        bigquery_schema = client.schema_from_json(tmp_schema_file.name)

        try:
            # destination table already exists, update schema
            table = client.get_table(destination_table)
            table.schema = bigquery_schema
            return client.update_table(table, ["schema"])
        except NotFound:
            table = bigquery.Table(destination_table, schema=bigquery_schema)
            return client.create_table(table)

    def merge(
        self,
        other: "Schema",
        exclude: Optional[List[str]] = None,
        add_missing_fields=True,
        attributes: Optional[List[str]] = None,
        ignore_incompatible_fields: bool = False,
        ignore_missing_fields: bool = False,
    ):
        """Merge another schema into the schema."""
        if "fields" in other.schema and "fields" in self.schema:
            self._traverse(
                "root",
                self.schema["fields"],
                other.schema["fields"],
                update=True,
                exclude=exclude,
                add_missing_fields=add_missing_fields,
                attributes=attributes,
                ignore_incompatible_fields=ignore_incompatible_fields,
                ignore_missing_fields=ignore_missing_fields,
            )

    def equal(self, other: "Schema") -> bool:
        """Compare to another schema."""
        try:
            self._traverse(
                "root", self.schema["fields"], other.schema["fields"], update=False
            )
            self._traverse(
                "root", other.schema["fields"], self.schema["fields"], update=False
            )
        except Exception as e:
            print(e)
            return False

        return True

    def compatible(self, other: "Schema") -> bool:
        """
        Check if schema is compatible with another schema.

        If there is a field missing in the schema that is part of the "other" schema,
        the schemas are still compatible. However, if there are fields missing in the
        "other" schema they are not compatible since, e.g. inserting data into the "other"
        schema that follows this schema would fail.
        """
        try:
            self._traverse(
                "root",
                self.schema["fields"],
                other.schema["fields"],
                update=False,
                ignore_missing_fields=True,
            )
            self._traverse(
                "root",
                other.schema["fields"],
                self.schema["fields"],
                update=False,
                ignore_missing_fields=False,
            )
        except Exception as e:
            print(e)
            return False

        return True

    @staticmethod
    def _node_with_mode(node):
        """Add default value for mode to node."""
        if "mode" in node:
            return node
        return {"mode": "NULLABLE", **node}

    def _traverse(
        self,
        prefix,
        columns,
        other_columns,
        update=False,
        add_missing_fields=True,
        ignore_missing_fields=False,
        exclude=None,
        attributes=None,
        ignore_incompatible_fields=False,
    ):
        """Traverses two schemas for validation and optionally updates the first schema."""
        nodes = {n["name"]: Schema._node_with_mode(n) for n in columns}
        other_nodes = {
            n["name"]: Schema._node_with_mode(n)
            for n in other_columns
            if exclude is None or n["name"] not in exclude
        }

        for node_name, node in other_nodes.items():
            field_path = node["name"] + (".[]" if node["mode"] == "REPEATED" else "")
            dtype = node["type"]

            if node_name in nodes:
                # node exists in schema, update attributes where necessary
                for node_attr_key, node_attr_value in node.items():
                    if attributes and node_attr_key not in attributes:
                        continue

                    if node_attr_key == "type":
                        # sometimes types have multiple names (e.g. INT64 and INTEGER)
                        # make it consistent here
                        node_attr_value = self._type_mapping.get(
                            node_attr_value, node_attr_value
                        )
                        nodes[node_name][node_attr_key] = self._type_mapping.get(
                            nodes[node_name][node_attr_key],
                            nodes[node_name][node_attr_key],
                        )

                    if node_attr_key not in nodes[node_name]:
                        if update:
                            # add field attributes if not exists in schema
                            nodes[node_name][node_attr_key] = node_attr_value
                            # Netlify has a problem starting 2022-03-07 where lots of
                            # logging slows down builds to the point where our builds hit
                            # the time limit and fail (bug 1761292), and this print
                            # statement accounts for 84% of our build logging.
                            # TODO: Uncomment this print when Netlify fixes the problem.
                            # print(
                            #    f"Attribute {node_attr_key} added to {prefix}.{field_path}"
                            # )
                        else:
                            if node_attr_key == "description":
                                print(
                                    "Warning: descriptions for "
                                    f"{prefix}.{field_path} differ"
                                )
                            else:
                                if not ignore_incompatible_fields:
                                    raise Exception(
                                        f"{node_attr_key} missing in {prefix}.{field_path}"
                                    )
                    elif nodes[node_name][node_attr_key] != node_attr_value:
                        # check field attribute diffs
                        if node_attr_key == "description":
                            # overwrite descripton for the "other" schema
                            print(
                                f"Warning: descriptions for {prefix}.{field_path} differ."
                            )
                        elif node_attr_key != "fields":
                            if not ignore_incompatible_fields:
                                raise Exception(
                                    f"Cannot merge schemas. {node_attr_key} attributes "
                                    f"for {prefix}.{field_path} are incompatible"
                                )

                if dtype == "RECORD" and nodes[node_name]["type"] == "RECORD":
                    # keep traversing nested fields
                    self._traverse(
                        f"{prefix}.{field_path}",
                        nodes[node_name]["fields"],
                        node["fields"],
                        update=update,
                        add_missing_fields=add_missing_fields,
                        ignore_missing_fields=ignore_missing_fields,
                        attributes=attributes,
                        ignore_incompatible_fields=ignore_incompatible_fields,
                    )
            else:
                if update and add_missing_fields:
                    # node does not exist in schema, add to schema
                    columns.append(node.copy())
                    print(f"Field {node_name} added to {prefix}")
                else:
                    if not ignore_missing_fields:
                        raise Exception(
                            f"Field {prefix}.{field_path} is missing in schema"
                        )

    def to_yaml_file(self, yaml_path: Path):
        """Write schema to the YAML file path."""
        with open(yaml_path, "w") as out:
            yaml.dump(self.schema, out, default_flow_style=False, sort_keys=False)

    def to_json_file(self, json_path: Path):
        """Write schema to the JSON file path."""
        with open(json_path, "w") as out:
            json.dump(self.schema["fields"], out, indent=2)

    def to_json(self):
        """Return the schema data as JSON."""
        return json.dumps(self.schema)

    def to_bigquery_schema(self) -> List[SchemaField]:
        """Get the BigQuery representation of the schema."""
        return [SchemaField.from_api_repr(field) for field in self.schema["fields"]]

    @classmethod
    def from_bigquery_schema(cls, fields: List[SchemaField]) -> "Schema":
        """Construct a Schema from the BigQuery representation."""
        return cls({"fields": [field.to_api_repr() for field in fields]})

    def generate_compatible_select_expression(
        self,
        target_schema: "Schema",
        fields_to_remove: Optional[Iterable[str]] = None,
        unnest_structs: bool = False,
        max_unnest_depth: int = 0,
        unnest_allowlist: Optional[Iterable[str]] = None,
    ) -> str:
        """Generate the select expression for the source schema based on the target schema.

        The output will include all fields of the target schema in the same order of the target.
        Any fields that are missing in the source schema are set to NULL.

        :param target_schema: The schema to coerce the current schema to.
        :param fields_to_remove: Given fields are removed from the output expression. Expressed as a
            list of strings with `.` separating each level of nesting, e.g. record_name.field.
        :param unnest_structs: If true, all record fields are expressed as structs with all nested
            fields explicitly listed. This allows the expression to be compatible even if the
            source schemas get new fields added. Otherwise, records are only unnested if they
            do not match the target schema.
        :param max_unnest_depth: Maximum level of struct nesting to explicitly unnest in
            the expression.
        :param unnest_allowlist: If set, only the given top-level structs are unnested.
        """

        def _type_info(node):
            """Determine the BigQuery type information from Schema object field."""
            dtype = node["type"]
            if dtype == "RECORD":
                dtype = (
                    "STRUCT<"
                    + ", ".join(
                        f"`{field['name']}` {_type_info(field)}"
                        for field in node["fields"]
                    )
                    + ">"
                )
            elif dtype == "FLOAT":
                dtype = "FLOAT64"
            if node.get("mode") == "REPEATED":
                return f"ARRAY<{dtype}>"
            return dtype

        def recurse_fields(
            _source_schema_nodes: List[Dict],
            _target_schema_nodes: List[Dict],
            path=None,
        ) -> str:
            if path is None:
                path = []

            select_expr = []
            source_schema_nodes = {n["name"]: n for n in _source_schema_nodes}
            target_schema_nodes = {n["name"]: n for n in _target_schema_nodes}

            # iterate through fields
            for node_name, node in target_schema_nodes.items():
                dtype = node["type"]
                node_path = path + [node_name]
                node_path_str = ".".join(node_path)

                if node_name in source_schema_nodes:  # field exists in app schema
                    # field matches, can query as-is
                    if node == source_schema_nodes[node_name] and (
                        # don't need to unnest scalar
                        dtype != "RECORD"
                        or not unnest_structs
                        # reached max record depth to unnest
                        or len(node_path) > max_unnest_depth > 0
                        # field not in unnest allowlist
                        or (
                            unnest_allowlist is not None
                            and node_path[0] not in unnest_allowlist
                        )
                    ):
                        if (
                            fields_to_remove is None
                            or node_path_str not in fields_to_remove
                        ):
                            select_expr.append(node_path_str)
                    elif (
                        dtype == "RECORD"
                    ):  # for nested fields, recursively generate select expression
                        if (
                            node.get("mode", None) == "REPEATED"
                        ):  # unnest repeated record
                            select_expr.append(
                                f"""
                                    ARRAY(
                                        SELECT
                                            STRUCT(
                                                {recurse_fields(
                                                    source_schema_nodes[node_name]['fields'],
                                                    node['fields'],
                                                    [node_name],
                                                )}
                                            )
                                        FROM UNNEST({node_path_str}) AS `{node_name}`
                                    ) AS `{node_name}`
                                """
                            )
                        else:  # select struct fields
                            select_expr.append(
                                f"""
                                    STRUCT(
                                        {recurse_fields(
                                            source_schema_nodes[node_name]['fields'],
                                            node['fields'],
                                            node_path,
                                        )}
                                    ) AS `{node_name}`
                                """
                            )
                    else:  # scalar value doesn't match, e.g. different types
                        select_expr.append(
                            f"CAST(NULL AS {_type_info(node)}) AS `{node_name}`"
                        )
                else:  # field not found in source schema
                    select_expr.append(
                        f"CAST(NULL AS {_type_info(node)}) AS `{node_name}`"
                    )

            return ", ".join(select_expr)

        return recurse_fields(
            self.schema["fields"],
            target_schema.schema["fields"],
        )

    def generate_select_expression(
        self,
        remove_fields: Optional[Iterable[str]] = None,
        unnest_structs: bool = False,
        max_unnest_depth: int = 0,
        unnest_allowlist: Optional[Iterable[str]] = None,
    ) -> str:
        """Generate the select expression for the schema which includes each field."""
        return self.generate_compatible_select_expression(
            self,
            remove_fields,
            unnest_structs,
            max_unnest_depth,
            unnest_allowlist,
        )



#!/usr/bin/env python3

"""
Import iProspect CSV data from moz-fx-data-marketing-prod-iprospect storage bucket.

The CSV files are updated daily and contain the last 30 days of data. This script
will import only CSV data of the file that was uploaded on the specified date into BigQuery.
The destination table will contain the data of all files, with submission_date
being used for partitioning and specifying when the CSV file was uploaded.

Historic data can change in more recent CSV files. The raw data will be used for
debugging and for reporting changes of specific values in
moz-fx-data-marketing-prod.iprospect.detail_export_diff_v1
"""

from argparse import ArgumentParser
from google.cloud import bigquery
from google.cloud import storage

import pandas as pd

parser = ArgumentParser(description=__doc__)
parser.add_argument("--date", required=True)  # expect string with format yyyy-mm-dd
parser.add_argument("--project", default="moz-fx-data-marketing-prod")
parser.add_argument("--bucket", default="moz-fx-data-marketing-prod-iprospect")
parser.add_argument("--prefix", default="mozilla_detail_export")
parser.add_argument("--dataset", default="iprospect")
parser.add_argument("--table", default="adspend_raw_v1")


def main():
    """Load CSV data to temporary table."""
    args = parser.parse_args()
    client = bigquery.Client(args.project)

    storage_client = storage.Client()
    blobs = list(
        storage_client.list_blobs(args.bucket, prefix=f"{args.prefix}_{args.date}")
    )

    if len(blobs) == 0:
        raise Exception(
            f"No iProspect data available for {args.date} in {args.bucket}/{args.prefix}"
        )

    uri = f"gs://{args.bucket}/{blobs[0].name}"
    df = pd.read_csv(uri)
    df["submission_date"] = pd.to_datetime(args.date)
    df["date"] = pd.to_datetime(df["date"]).dt.date

    job_config = bigquery.LoadJobConfig(
        write_disposition="WRITE_TRUNCATE",
        time_partitioning=bigquery.TimePartitioning(
            type_=bigquery.TimePartitioningType.DAY,
            field="submission_date",
        ),
        schema=[
            bigquery.SchemaField("date", "DATE"),
            bigquery.SchemaField("fetch_ad_name", "STRING"),
            bigquery.SchemaField("vendor", "STRING"),
            bigquery.SchemaField("campaign", "STRING"),
            bigquery.SchemaField("channel", "STRING"),
            bigquery.SchemaField("country", "STRING"),
            bigquery.SchemaField("creative", "STRING"),
            bigquery.SchemaField("creative_concept", "STRING"),
            bigquery.SchemaField("creative_language", "STRING"),
            bigquery.SchemaField("creative_size", "STRING"),
            bigquery.SchemaField("creative_type", "STRING"),
            bigquery.SchemaField("device", "STRING"),
            bigquery.SchemaField("goal", "STRING"),
            bigquery.SchemaField("media_type", "STRING"),
            bigquery.SchemaField("operating_system", "STRING"),
            bigquery.SchemaField("placement", "STRING"),
            bigquery.SchemaField("product", "STRING"),
            bigquery.SchemaField("social_string", "STRING"),
            bigquery.SchemaField("targeting", "STRING"),
            bigquery.SchemaField("traffic_type", "STRING"),
            bigquery.SchemaField("imps_vendor", "INT64"),
            bigquery.SchemaField("clicks_vendor", "INT64"),
            bigquery.SchemaField("spend_vendor", "FLOAT64"),
            bigquery.SchemaField("client_fee", "FLOAT64"),
            bigquery.SchemaField("client_cost", "FLOAT64"),
            bigquery.SchemaField("video_3sec_vendor", "INT64"),
            bigquery.SchemaField("video_completions_vendor", "INT64"),
            bigquery.SchemaField("video_firstquartile_vendor", "INT64"),
            bigquery.SchemaField("video_midpoint_vendor", "INT64"),
            bigquery.SchemaField("video_thirdquartile_vendor", "INT64"),
            bigquery.SchemaField("video_views_vendor", "INT64"),
            bigquery.SchemaField("conversions", "FLOAT64"),
            bigquery.SchemaField("post_click_conversions", "FLOAT64"),
            bigquery.SchemaField("post_view_conversions", "FLOAT64"),
            bigquery.SchemaField("tweet_engagements", "FLOAT64"),
            bigquery.SchemaField("bounces_ga", "INT64"),
            bigquery.SchemaField("new_users_ga", "INT64"),
            bigquery.SchemaField("firefox_downloads_ga", "INT64"),
            bigquery.SchemaField("sessions_ga", "INT64"),
            bigquery.SchemaField("firefox_downloads_goal_2_sa360", "INT64"),
            bigquery.SchemaField("firefox_downloads_sa360", "INT64"),
            bigquery.SchemaField("desktop_downloads_sa360", "INT64"),
            bigquery.SchemaField("adjust_clicks", "INT64"),
            bigquery.SchemaField("adjust_installs", "INT64"),
            bigquery.SchemaField("adjust_sessions", "INT64"),
            bigquery.SchemaField("adjust_organic_clicks", "INT64"),
            bigquery.SchemaField("adjust_organic_installs", "INT64"),
            bigquery.SchemaField("adjust_organic_sessions", "INT64"),
            bigquery.SchemaField("dcm_active_view_eligible_impressions", "INT64"),
            bigquery.SchemaField("dcm_active_view_measurable_impressions", "INT64"),
            bigquery.SchemaField("dcm_active_view_viewable_impressions", "INT64"),
            bigquery.SchemaField("submission_date", "DATE"),
        ],
    )

    partition = args.date.replace("-", "")
    destination = f"{args.project}.{args.dataset}.{args.table}${partition}"
    job = client.load_table_from_dataframe(df, destination, job_config=job_config)

    print(f"Running job {job.job_id}")
    job.result()
    print(f"Loaded {uri} for {args.date}")


if __name__ == "__main__":
    main()



"""Update metadata of BigQuery tables and views."""

from argparse import ArgumentParser
from pathlib import Path

from google.cloud import bigquery

from ..config import ConfigLoader
from ..util import standard_args
from .parse_metadata import ExternalDataFormat, Metadata

METADATA_FILE = "metadata.yaml"
DEFAULT_PATTERN = (
    f"{ConfigLoader.get('default', 'project', fallback='moz-fx-data-shared-prod')}:*.*"
)


parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "patterns",
    metavar="[project:]dataset[.table]",
    default=[DEFAULT_PATTERN],
    nargs="*",
    help="Table that should have a latest-version view, may use shell-style wildcards,"
    f" defaults to: {DEFAULT_PATTERN}",
)
parser.add_argument("--target", help="File or directory containing metadata files")
standard_args.add_log_level(parser)


class InvalidExternalDataConfigException(Exception):
    """Raised invalid config for external data tables."""


def publish_metadata(client, project, dataset, table, metadata):
    """Push metadata to BigQuery tables."""
    try:
        table_ref = client.dataset(dataset).table(table)
        table = client.get_table(table_ref)

        if metadata.friendly_name is not None:
            table.friendly_name = metadata.friendly_name

        if metadata.description is not None:
            table.description = metadata.description

        table.labels = {
            key: value
            for key, value in metadata.labels.items()
            if isinstance(value, str)
        }

        if metadata.deprecated is True:
            table.labels["deprecated"] = "true"
        if metadata.deletion_date:
            table.labels["deletion_date"] = metadata.deletion_date.strftime("%Y-%m-%d")
            # TODO: in the future we can consider updating the table expiration date based on deletion_date
        if metadata.monitoring and metadata.monitoring.enabled:
            table.labels["monitoring"] = "true"

        client.update_table(table, ["friendly_name", "description", "labels"])
        print("Published metadata for: {}.{}.{}".format(project, dataset, table))

    except Exception as e:
        print(e)


def attach_metadata(artifact_file_path: Path, table: bigquery.Table) -> None:
    """Add metadata from query file's metadata.yaml to table object."""
    try:
        if artifact_file_path.is_file() and artifact_file_path.name == METADATA_FILE:
            metadata = Metadata.from_file(artifact_file_path)
        else:
            metadata = Metadata.of_query_file(artifact_file_path)
    except FileNotFoundError:
        return

    table.description = metadata.description
    table.friendly_name = metadata.friendly_name

    if metadata.bigquery and metadata.bigquery.time_partitioning:
        table.time_partitioning = bigquery.TimePartitioning(
            metadata.bigquery.time_partitioning.type.bigquery_type,
            field=metadata.bigquery.time_partitioning.field,
            require_partition_filter=(
                metadata.bigquery.time_partitioning.require_partition_filter
            ),
            expiration_ms=metadata.bigquery.time_partitioning.expiration_ms,
        )
    elif metadata.bigquery and metadata.bigquery.range_partitioning:
        table.range_partitioning = bigquery.RangePartitioning(
            field=metadata.bigquery.range_partitioning.field,
            range_=bigquery.PartitionRange(
                start=metadata.bigquery.range_partitioning.range.start,
                end=metadata.bigquery.range_partitioning.range.end,
                interval=metadata.bigquery.range_partitioning.range.interval,
            ),
        )

    if metadata.bigquery and metadata.bigquery.clustering:
        table.clustering_fields = metadata.bigquery.clustering.fields

    # BigQuery only allows for string type labels with specific requirements to be published:
    # https://cloud.google.com/bigquery/docs/labels-intro#requirements
    if metadata.labels:
        table.labels = {
            key: value
            for key, value in metadata.labels.items()
            if isinstance(value, str)
        }


def attach_external_data_config(artifact_file_path, table) -> None:
    """Add external data metadata from query file's metadata.yaml to table object."""
    try:
        if artifact_file_path.is_file() and artifact_file_path.name == METADATA_FILE:
            metadata = Metadata.from_file(artifact_file_path)
        else:
            metadata = Metadata.of_query_file(artifact_file_path)
    except FileNotFoundError:
        raise InvalidExternalDataConfigException(
            f"Invalid metadata: External data table "
            f"{artifact_file_path} missing metadata file"
        )

    if not metadata.external_data:
        raise InvalidExternalDataConfigException(
            f"Invalid metadata: External data table "
            f"{artifact_file_path} has no external_data config"
        )

    if metadata.external_data.format not in (
        ExternalDataFormat.GOOGLE_SHEETS,
        ExternalDataFormat.CSV,
    ):
        raise InvalidExternalDataConfigException(
            f"Invalid metadata: External data table "
            f"{artifact_file_path} has unsupported format {metadata.external_data.format}"
        )

    external_config = bigquery.ExternalConfig(
        metadata.external_data.format.value.upper()
    )
    external_config.source_uris = metadata.external_data.source_uris
    external_config.ignore_unknown_values = True
    external_config.autodetect = False

    for key, v in metadata.external_data.options.items():
        setattr(external_config.options, key, v)

    table.external_data_configuration = external_config



#!/usr/bin/env python3

"""Generates the list of missing columns in structured datasets."""
import datetime
import json
from argparse import ArgumentParser
from collections import defaultdict

from google.cloud import bigquery

NAMESPACE_QUERY = r"""
SELECT
  schema_name
FROM
  `moz-fx-data-shared-prod`.INFORMATION_SCHEMA.SCHEMATA
WHERE
  schema_name LIKE r"%\_stable"
  AND schema_name NOT LIKE "%telemetry%"
ORDER BY
  schema_name
"""

QUERY = """
-- missing columns {date} {namespace};
WITH extracted AS (
  SELECT
    DATE "{date}" AS submission_date,
    "{namespace}" AS document_namespace,
    `moz-fx-data-shared-prod`.udf.extract_document_type(_TABLE_SUFFIX) AS document_type,
    `moz-fx-data-shared-prod`.udf.extract_document_version(_TABLE_SUFFIX) AS document_version,
    additional_properties
  FROM
    `moz-fx-data-shared-prod.{namespace}_stable.*`
  WHERE
    -- only observe full days of data
    DATE(submission_timestamp) = date "{date}"
    -- https://cloud.google.com/bigquery/docs/querying-wildcard-tables#filtering_selected_tables_using_table_suffix
    -- IT's also possible to exclude tables in this query e.g
    -- AND _TABLE_SUFFIX NOT IN ('main_v5', 'saved_session_v5', 'first_shutdown_v5')
),
transformed AS (
  SELECT
    * EXCEPT (additional_properties),
    COUNT(*) AS path_count
  FROM
    extracted,
    UNNEST(
      `moz-fx-data-shared-prod`.udf_js.json_extract_missing_cols(
        additional_properties,
        [],
        -- TODO: manually curate known missing schema bits
        []
      )
    ) AS path
  GROUP BY
    submission_date,
    document_namespace,
    document_type,
    document_version,
    path
)
SELECT
  submission_date,
  document_namespace,
  document_type,
  document_version,
  path,
  path_count
FROM
  transformed
"""


def structured_missing_columns(
    date, project, destination_dataset, destination_table, skip_exceptions
):
    """Get missing columns for structured ingestion."""
    client = bigquery.Client(project=project)
    print("Getting the list of namespaces...")
    namespaces = [
        row.schema_name.split("_stable")[0]
        for row in client.query(NAMESPACE_QUERY).result()
    ]
    print(namespaces)

    mb_processed = {}
    for i, namespace in enumerate(namespaces):
        print(f"Running job for {namespace}")
        try:
            job = client.query(
                QUERY.format(date=date, namespace=namespace),
                job_config=bigquery.QueryJobConfig(
                    time_partitioning=bigquery.TimePartitioning(
                        field="submission_date"
                    ),
                    create_disposition=bigquery.CreateDisposition.CREATE_IF_NEEDED,
                    write_disposition=(
                        # only the first query should overwrite the partition
                        bigquery.WriteDisposition.WRITE_TRUNCATE
                        if i == 0
                        else bigquery.WriteDisposition.WRITE_APPEND
                    ),
                    destination=(
                        f"{project}.{destination_dataset}.{destination_table}"
                        f"${date.strftime('%Y%m%d')}"
                    ),
                ),
            )
            job.result()
            mb_processed[namespace] = round(job.total_bytes_processed / 1024**2, 1)
        except Exception as e:
            if not skip_exceptions:
                raise
            print(e)
    # summary info
    print(json.dumps(mb_processed, indent=2))


def parse_args():
    """Parse command line arguments."""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("--date", type=datetime.date.fromisoformat, required=True)
    parser.add_argument("--project", default="moz-fx-data-shared-prod")
    parser.add_argument("--destination-dataset", default="monitoring_derived")
    parser.add_argument("--destination-table", default="structured_missing_columns_v1")
    parser.add_argument(
        "--skip-exceptions",
        action="store_true",
        default=False,
        help="continue when running into errors during query time",
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    structured_missing_columns(
        args.date,
        args.project,
        args.destination_dataset,
        args.destination_table,
        args.skip_exceptions,
    )



"""Upload symbols used in crash pings."""

from datetime import datetime
from functools import partial
from multiprocessing.pool import ThreadPool as Pool
from pathlib import Path

import click
import requests
import symbolic
import yaml
from google.api_core.exceptions import BadRequest, NotFound
from google.cloud import bigquery
from requests.adapters import HTTPAdapter
from requests.exceptions import ConnectionError, HTTPError
from urllib3.util import Retry

SCHEMA_FILE = Path(__file__).parent / "schema.yaml"
OUTPUT_SCHEMA = bigquery.SchemaField.from_api_repr(
    {"name": "root", "type": "RECORD", **yaml.safe_load(SCHEMA_FILE.read_text())}
).fields


def _bytes_split_generator(item, sep):
    # https://github.com/mozilla-services/eliot/blob/be74cd6c0ef09dd85a71c8b1b22c3297b6b9f9bf/eliot/libsymbolic.py#L18-L36
    index = 0
    len_item = len(item)
    while index <= len_item:
        next_index = item.find(sep, index)
        if next_index == -1:
            break
        yield item[index:next_index]
        index = next_index + len(sep)


def _get_module_filename(sym_file, debug_filename):
    # https://github.com/mozilla-services/eliot/blob/be74cd6c0ef09dd85a71c8b1b22c3297b6b9f9bf/eliot/libsymbolic.py#L39-L68
    # Iterate through the first few lines of the file until we hit FILE in which
    # case there's no INFO for some reason or we hit the first INFO.
    for line in _bytes_split_generator(sym_file, b"\n"):
        if line.startswith(b"INFO"):
            parts = line.split(b" ")
            if len(parts) == 4:
                return parts[-1].decode("utf-8").strip()
            else:
                break
        elif line.startswith((b"FILE", b"PUBLIC", b"FUNC")):
            break
    return debug_filename


def _get_symbols(session, source_url, submission_date, row):
    """Get symbols needed from a single symbol file.

    This may be rearchitected to use eliot by sending per-module jobs.
    """
    debug_file = row["debug_file"]
    debug_id = row["debug_id"]
    module_offsets = row["module_offsets"]

    symbols = []
    if debug_file.endswith(".pdb"):
        sym_filename = debug_file[:-4] + ".sym"
    else:
        sym_filename = debug_file + ".sym"

    request_timestamp = datetime.utcnow()
    try:
        resp = session.get(
            f"{source_url}/{debug_file}/{debug_id}/{sym_filename}",
            allow_redirects=True,
        )
        if resp.status_code not in (400, 404):
            resp.raise_for_status()
    except (ConnectionError, HTTPError) as e:
        print("ERROR: could not get symbols: " f"{debug_file} {debug_id} {e}")
        return e, []
    if resp.status_code == 404:
        print("WARNING: symbols not found: " f"{debug_file} {debug_id}")
        return None, []
    if resp.status_code == 400:
        print(
            f"400 Invalid Request {resp.request.url}: "
            f"{resp.request.body.decode('utf-8')!r}"
        )
        resp.raise_for_status()

    sym_file = resp.content

    norm_debug_id = symbolic.normalize_debug_id(debug_id)
    sym_archive = symbolic.Archive.from_bytes(sym_file)
    symcache = sym_archive.get_object(debug_id=norm_debug_id).make_symcache()
    module_filename = _get_module_filename(sym_file, debug_file)

    # https://github.com/mozilla-services/eliot/blob/be74cd6c0ef09dd85a71c8b1b22c3297b6b9f9bf/eliot/symbolicate_resource.py#L505-L553
    for module_offset in module_offsets:
        frame_info = {
            "submission_date": submission_date,
            "request_timestamp": request_timestamp.isoformat(),
            "debug_file": debug_file,
            "debug_id": debug_id,
            "module_offset": hex(module_offset),
            "module": module_filename,
        }

        if module_offset < 0:
            continue  # ignore invalid offset

        sourceloc_list = symcache.lookup(module_offset)
        if not sourceloc_list:
            continue  # no symbols for this offset

        # sourceloc_list can have multiple entries: It starts with the innermost
        # inline stack frame, and then advances to its caller, and then its
        # caller, and so on, until it gets to the outer function.
        # We process the outer function first, and then add inline stack frames
        # afterwards. The outer function is the last item in sourceloc_list.
        sourceloc = sourceloc_list[-1]

        frame_info["function"] = sourceloc.symbol
        frame_info["function_offset"] = hex(module_offset - sourceloc.sym_addr)
        if sourceloc.full_path:
            frame_info["file"] = sourceloc.full_path

            # Only add a "line" if it's non-zero and not None, and if there's a
            # file--otherwise the line doesn't mean anything
            if sourceloc.line:
                frame_info["line"] = sourceloc.line

        if len(sourceloc_list) > 1:
            # We have inline information. Add an "inlines" property with a list
            # of { function, file, line } entries.
            inlines = []
            for inline_sourceloc in sourceloc_list[:-1]:
                inline_data = {
                    "function": inline_sourceloc.symbol,
                }

                if inline_sourceloc.full_path:
                    inline_data["file"] = inline_sourceloc.full_path

                    if inline_sourceloc.line:
                        inline_data["line"] = inline_sourceloc.line

                inlines.append(inline_data)

            frame_info["inlines"] = inlines
        symbols.append(frame_info)
    return None, symbols


@click.command()
@click.option(
    "--collect-only-missing/--collect-all",
    default=True,
    help="Collect only missing symbols or collect all symbols for --submission-date",
)
@click.option(
    "--parallelism",
    default=40,
    help="Number of threads to use when downloading symbols."
    " Default assumes at least 100GiB of ram available.",
)
@click.option(
    "--submission-date",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    required=True,
)
@click.option(
    "--source-table",
    default="moz-fx-data-shared-prod.telemetry_derived.crash_frames_v1",
)
@click.option(
    "--source-url",
    default="https://symbols.mozilla.org",
)
@click.option(
    "--destination-table",
    default="moz-fx-data-shared-prod.telemetry_derived.crash_symbols_v1",
)
def main(
    collect_only_missing,
    parallelism,
    submission_date,
    source_table,
    source_url,
    destination_table,
):
    bq = bigquery.Client()
    query_job = bq.query(
        f"""
        SELECT
          debug_file,
          debug_id,
          ARRAY_AGG(DISTINCT module_offset ORDER BY module_offset) AS module_offsets,
        FROM
          `{source_table}`
        WHERE
          DATE(submission_timestamp) = @submission_date
          AND debug_file != ""
          AND debug_id != ""
        GROUP BY
          debug_file,
          debug_id
        """,
        job_config=bigquery.QueryJobConfig(
            query_parameters=[
                bigquery.ScalarQueryParameter(
                    "submission_date", "DATE", f"{submission_date:%F}"
                ),
            ],
        ),
    )
    rows = [{**row} for row in query_job.result()]

    if collect_only_missing:
        # don't collect symbols already present in the destination table
        try:
            bq.get_table(destination_table)
        except NotFound:
            pass  # collect all symbols
        else:
            existing_symbols_job = bq.query(
                f"""
                SELECT
                  debug_id,
                  debug_file,
                  ARRAY_AGG(DISTINCT module_offset) AS module_offsets,
                FROM
                  `{destination_table}`
                WHERE
                  submission_date = @submission_date
                GROUP BY
                  debug_id,
                  debug_file
                """,
                job_config=bigquery.QueryJobConfig(
                    query_parameters=[
                        bigquery.ScalarQueryParameter(
                            "submission_date", "DATE", f"{submission_date:%F}"
                        ),
                    ],
                ),
            )
            existing_symbols = {
                (row["debug_file"], row["debug_id"]): set(row["module_offsets"])
                for row in existing_symbols_job.result()
            }
            for row in rows:
                existing_module_offsets = existing_symbols.get(
                    (row["debug_file"], row["debug_id"])
                )
                if existing_module_offsets is not None:
                    row["module_offsets"] = [
                        o
                        for o in row["module_offsets"]
                        if o not in existing_module_offsets
                    ]
            rows = [row for row in rows if row["module_offsets"]]  # drop empty rows

    retry_strategy = Retry(
        total=4,
        status_forcelist=[429, 500, 502, 503, 504],
    )
    adapter = HTTPAdapter(max_retries=retry_strategy)
    session = requests.Session()
    session.mount("https://", adapter)
    _get_symbols_with_args = partial(_get_symbols, session, source_url, submission_date)

    symbols = []
    symbol_errors = []
    with Pool(parallelism) as pool:
        for error, _symbols in pool.imap(_get_symbols_with_args, rows, chunksize=1):
            if error is None:
                symbols.extend(symbols)
            else:
                symbol_errors.append(error)

    load_job = bq.load_table_from_json(
        json_rows=symbols,
        destination=f"{destination_table}${submission_date:%Y%m%d}",
        job_config=bigquery.LoadJobConfig(
            ignore_unknown_values=False,
            schema=OUTPUT_SCHEMA,
            write_disposition=bigquery.WriteDisposition.WRITE_APPEND,
        ),
    )
    try:
        load_job.result()
    except BadRequest as e:
        print(e.errors)
        print(load_job.errors)
        raise

    if symbol_errors:
        # errors collecting symbols are deferred to here so that retries
        # only need to collect symbols that previously failed
        raise Exception(
            "failed to collect some symbols, "
            "run again in append mode to retry only missing symbols"
        )


if __name__ == "__main__":
    main()
